import tempfile, shutil
from urllib.request import urlopen
from coldtype.renderable import renderable
from coldtype.webserver.cdelopty import evalcdel
from coldtype.pens.datpen import DATPen, DATPens
from coldtype.pens.skiapen import SkiaPen
from coldtype.text.reader import Font #, StyledString, Style
from coldtype.geometry import Rect
from coldtype.color import hsl
from flask import send_file

cottf_path = "https://blog.robstenson.com/assets/ColdtypeObviously-VF.ttf"

with urlopen(cottf_path) as response:
    with tempfile.NamedTemporaryFile(suffix=".ttf", delete=False) as tf:
        shutil.copyfileobj(response, tf)
        print(tf.name)

cottf = Font(tf.name)
fonts = {"co": cottf}

test = [
    ["P", "®", ".", ["f", 0]],
    ["P", ".",
        ["oval", ["R", "®", ".", ["inset", 50]]],
        ["f", ["hsl", 0.6, {"s":1}]],
        ["align", "®"]],
    ["S", "COLD", "co", 500,
        {"wdth":0.5, "tu":-80, "r":1, "ro": 1, "rotate":-5},
        ".",
        ["pens"],
        ["align", "®"],
        ["f", 1],
        ["understroke", 0, 30],
        ["rotate", 15]]]

DATPen._pen_class = SkiaPen

@renderable(Rect(1080, 1080), rstate=1)
def test1(r, get):
    return evalcdel(get.get("cdel", test), r, fonts)

def hello_world(req):
    rp = test1.passes(None, None)[0]
    result = test1.normalize_result(test1.run(rp, req.args))
    with tempfile.NamedTemporaryFile(suffix="png") as tf:
        SkiaPen.Composite(result, test1.rect, tf.name, context=None)
        return send_file(tf.name, as_attachment=False, attachment_filename="test.png")

if __name__ == "__main__":
    from flask import Flask, request
    app = Flask(__name__)

    @app.route('/')
    def test_hello_world():
        return hello_world(request)
    
    app.run(debug=True)


import importlib
import shlex
import signal
import sys
import traceback


def ignoreSignal(sig, frame):
    pass


def raiseKeyboardInterrupt(sig, frame):
    raise KeyboardInterrupt()


ERROR_MARKER = "---- ERROR ----"
SUCCESS_MARKER = "---- SUCCESS ----"


def workServer():
    signal.signal(signal.SIGINT, ignoreSignal)
    while True:
        input = sys.stdin.readline()
        input = input.strip()
        if not input:
            break
        try:
            try:
                signal.signal(signal.SIGINT, raiseKeyboardInterrupt)
                command, *args = shlex.split(input)
                moduleName, funcName = command.rsplit(".", 1)
                module = importlib.import_module(moduleName)
                func = getattr(module, funcName)
                func(*args)
            finally:
                signal.signal(signal.SIGINT, ignoreSignal)
        except KeyboardInterrupt:
            print(ERROR_MARKER)
        except:
            traceback.print_exc()
            print(ERROR_MARKER)
        else:
            print(SUCCESS_MARKER)


if __name__ == "__main__":
    workServer()



from coldtype.geometry import Rect, Edge, Point
from coldtype.beziers import CurveCutter, raise_quadratic
from coldtype.pens.drawablepen import DrawablePenMixin
from fontTools.pens.basePen import BasePen
from coldtype.color import Gradient, Color, normalize_color

"""
import math
from random import random, randint, seed

from coldtype.text.reader import StyledString, Style
from coldtype.pens.blenderpen import BlenderPen, BPH
from coldtype.geometry import Rect
from coldtype.pens.draftingpen import DraftingPen

BPH.Clear()

r = Rect(0, 0, 1000, 1000)
tc = BPH.Collection("Text")

DraftingPen().f(1).rect(r).tag("Frame").cast(BlenderPen).draw(tc, plane=1)

StyledString("Yoy", Style("~/Type/fonts/fonts/Nonplus_Black.otf", 200, fill=(1, 0, 0.5))).pen().align(r).tag("Yoy").cast(BlenderPen).draw(tc)
"""

import math
import random
try:
    # Blender-specific things
    import bpy
    from mathutils import Vector, Matrix
except:
    print(">>> Not a blender environment")


class BPH():
    def Clear():
        print(">>>CLERAING")
        for block in bpy.data.meshes:
            if block.users == 0:
                bpy.data.meshes.remove(block)

        for block in bpy.data.materials:
            if block.users == 0:
                bpy.data.materials.remove(block)

        for block in bpy.data.textures:
            if block.users == 0:
                bpy.data.textures.remove(block)

        for block in bpy.data.images:
            if block.users == 0:
                bpy.data.images.remove(block)

    def FindCollectionForItem(item):
        collections = item.users_collection
        if len(collections) > 0:
            return collections[0]
        return bpy.context.scene.collection

    def Collection(name, parent=None):
        if name not in bpy.data.collections:
            coll = bpy.data.collections.new(name)
            if parent:
                parent.children.link(coll)
            else:
                bpy.context.scene.collection.children.link(coll)
        return bpy.data.collections.get(name)

    def Primitive(_type, coll, name, dn=False, container=None):
        created = False
        
        if dn and name in bpy.context.scene.objects:
            obj = bpy.context.scene.objects[name]
            bpy.data.objects.remove(obj, do_unlink=True)

        if name not in bpy.context.scene.objects:
            created = True
            if _type == "Bezier":
                bpy.ops.curve.primitive_bezier_curve_add()
            elif _type == "Plane":
                bpy.ops.mesh.primitive_plane_add()
            bc = bpy.context.active_object
            bc.name = name
            bc.data.name = name
            if _type == "Bezier":
                bc.data.dimensions = "2D"
                bc.data.fill_mode = "BOTH"
                bc.data.extrude = 0.1
            elif _type == "Plane":
                if container:
                    bc.scale[0] = container.w/2
                    bc.scale[1] = container.h/2
                    bc.location[0] = container.x + container.w/2
                    bc.location[1] = container.y + container.h/2
                    bpy.ops.object.origin_set(type="ORIGIN_CURSOR")
                    bpy.ops.object.transform_apply()
            mat = bpy.data.materials.new(f"Material_{name}")
            mat.use_nodes = True
            bc.data.materials.append(mat)
        
        bc = bpy.context.scene.objects[name]
        bc_coll = BPH.FindCollectionForItem(bc)
        if bc_coll != coll:
            coll.objects.link(bc)
            bc_coll.objects.unlink(bc)
        bc.select_set(False)

        return bc, created
    
    def Vector(pt, z=0):
        x, y = pt
        return Vector((x, y, z))


class BlenderPen(DrawablePenMixin, BasePen):
    def __init__(self, dat):
        super().__init__(None)
        self.dat = dat
        tag = self.dat.tag()
        self.tag = tag if tag and tag != "Unknown" else f"Curve_{random.randint(0, 1000000)}"
    
    def record(self, dat):
        self._spline = None
        self.splines = []
        self._value = []
        dat.replay(self)
        if self._spline and len(self._spline) > 0 and self._spline not in self.splines:
            self.splines.append(self._spline)

    def _moveTo(self, p):
        if self._spline and len(self._spline) > 0 and self._spline not in self.splines:
            self.splines.append(self._spline)
        self._spline = []
        self._value.append([p])
        self._spline.append(["BEZIER", "start", [p, p, p]])

    def _lineTo(self, p):
        self._value.append([p])
        self._spline.append(["BEZIER", "curve", [p, p, p]])

    def _curveToOne(self, p1, p2, p3):
        self._spline[-1][-1][-1] = p1
        self._value.append([p1, p2, p3])
        self._spline.append(["BEZIER", "curve", [p2, p3, p3]])

    def _qCurveToOne(self, p1, p2):
        start = self._value[-1][-1]
        q1, q2, q3 = raise_quadratic(start, (p1[0], p1[1]), (p2[0], p2[1]))
        self._spline[-1][-1][-1] = q1
        self._value.append([q1, q2, q3])
        self._spline.append(["BEZIER", "curve", [q2, q3, q3]])

    def _closePath(self):
        if self._spline and len(self._spline) > 0 and self._spline not in self.splines:
            self.splines.append(self._spline)
            self.spline = None
        self.spline = None
    
    def materials(self):
        return self.bez.data.materials

    def bsdf(self):
        return self.materials()[0].node_tree.nodes["Principled BSDF"]
    
    def shadow(self, clip=None, radius=10, alpha=0.3, color=Color.from_rgb(0,0,0,1)):
        pass

    def setColorValue(self, value, color):
        value[0] = color.r
        value[1] = color.g
        value[2] = color.b
        value[3] = 1
        if color.a != 1:
            self.transmission(1)
            #value[0] = 1
            #value[1] = 1
            #value[2] = 1
            #value[3] = 1
    
    def fill(self, color):
        if color:
            if isinstance(color, Gradient):
                self.fill(color.stops[0][0])
            else:
                #print("FILL>>>>>", self.tag, color)
                bsdf = self.bsdf()
                dv = bsdf.inputs[0].default_value
                self.setColorValue(dv, color)
    
    def stroke(self, weight=1, color=None, dash=None):
        if weight and color and color.a > 0:
            #print("STROKE>>>", self.tag, weight, color)
            self.bez.data.fill_mode = "NONE"
            if isinstance(color, Gradient):
                pass
            else:
                self.fill(color)
            
    def extrude(self, amount=0.1):
        self.bez.data.extrude = amount
        return self
    
    def bevel(self, depth=0.02):
        self.bez.data.bevel_depth = depth
        return self
    
    def metallic(self, amount=1):
        self.bsdf().inputs[4].default_value = amount
        return self
    
    def transmission(self, amount=1):
        self.bsdf().inputs[15].default_value = amount
        return self
    
    def subsurface(self, amount=0.01):
        self.bsdf().inputs[1].default_value = amount
        return self

    def emission(self, color=None):
        if color is not None:
            self.setColorValue(self.bsdf().inputs[17].default_value, normalize_color(color))
        return self
    
    def image(self, path):
        mat = self.materials()[0]
        bsdf = self.bsdf()
        if "Image Texture" in mat.node_tree.nodes:
            imgtex = mat.node_tree.nodes["Image Texture"]
        else:
            imgtex = mat.node_tree.nodes.new("ShaderNodeTexImage")
            mat.node_tree.links.new(bsdf.inputs["Base Color"], imgtex.outputs["Color"])
        imgtex.image = bpy.data.images.load(path)
        return self
    
    def convertToMesh(self):
        self.bez.select_set(True)
        bpy.ops.object.convert(target="MESH")
        self.bez.select_set(False)
        return self
    
    def solidify(self, thickness=1):
        bpy.context.view_layer.objects.active = None
        bpy.context.view_layer.objects.active = self.bez
        self.bez.select_set(True)
        if "Solidify" not in self.bez.modifiers:
            bpy.ops.object.modifier_add(type="SOLIDIFY")
        self.bez.modifiers["Solidify"].thickness = thickness
        self.bez.select_set(False)
        bpy.context.view_layer.objects.active = None
        return self
    
    def applyModifier(self):
        bpy.context.view_layer.objects.active = None
        bpy.context.view_layer.objects.active = self.bez
        self.bez.select_set(True)
        bpy.ops.object.modifier_apply(apply_as="DATA")
        self.bez.select_set(False)
        bpy.context.view_layer.objects.active = None
        return self
    
    def rotate(self, x=None, y=None, z=None):
        if x is not None:
            self.bez.rotation_euler[0] = math.radians(x)
        if y is not None:
            self.bez.rotation_euler[1] = math.radians(y)
        if z is not None:
            self.bez.rotation_euler[2] = math.radians(z)
        return self
    
    def locate(self, x=None, y=None, z=None):
        if x is not None:
            self.bez.location[0] = x
        if y is not None:
            self.bez.location[1] = y
        if z is not None:
            self.bez.location[2] = z
        return self
    
    def draw(self, collection, style=None, scale=0.01, cyclic=True, dn=False, plane=False):
        if plane:
            self.bez, self.created = BPH.Primitive("Plane", collection, self.tag, dn=dn, container=self.dat.ambit().scale(scale))
        else:
            self.bez, self.created = BPH.Primitive("Bezier", collection, self.tag, dn=dn)
            self.bez.data.fill_mode = "BOTH"
            self.record(self.dat.copy().removeOverlap().scale(scale, point=False))
            self.drawOnBezierCurve(self.bez.data, cyclic=cyclic)
        for attrs, attr in self.findStyledAttrs(style):
            self.applyDATAttribute(attrs, attr)
        return self

    def drawOnBezierCurve(self, bez, cyclic=True):
        for spline in reversed(bez.splines): # clear existing splines
            bez.splines.remove(spline)

        for spline_data in self.splines:
            bez.splines.new('BEZIER')
            spline = bez.splines[-1]
            spline.use_cyclic_u = cyclic
            for i, (t, style, pts) in enumerate(spline_data):
                l, c, r = pts
                if i > 0:
                    spline.bezier_points.add(1)
                pt = spline.bezier_points[-1]
                pt.co = BPH.Vector(c)
                pt.handle_left = BPH.Vector(l)
                pt.handle_right = BPH.Vector(r)



import skia, struct, tempfile

from pathlib import Path

from fontTools.pens.transformPen import TransformPen
from fontTools.pens.basePen import BasePen

from coldtype.pens.datpen import DATPen
from coldtype.pens.dattext import DATText
from coldtype.img.datimage import DATImage
from coldtype.geometry import Rect, Point
from coldtype.pens.drawablepen import DrawablePenMixin, Gradient
from coldtype.color import Color
from coldtype.text.reader import Style


class SkiaPathPen(BasePen):
    def __init__(self, dat, h):
        super().__init__()
        self.dat = dat
        self.path = skia.Path()

        tp = TransformPen(self, (1, 0, 0, -1, 0, h))
        dat.replay(tp)
    
    def _moveTo(self, p):
        self.path.moveTo(p[0], p[1])

    def _lineTo(self, p):
        self.path.lineTo(p[0], p[1])

    def _curveToOne(self, p1, p2, p3):
        self.path.cubicTo(p1[0], p1[1], p2[0], p2[1], p3[0], p3[1])

    def _qCurveToOne(self, p1, p2):
        self.path.quadTo(p1[0], p1[1], p2[0], p2[1])

    def _closePath(self):
        self.path.close()


class SkiaPen(DrawablePenMixin, SkiaPathPen):
    def __init__(self, dat, rect, canvas, scale, style=None, alpha=1):
        super().__init__(dat, rect.h)
        self.scale = scale
        self.canvas = canvas
        self.rect = rect
        self.blendmode = None
        self.style = style
        self.alpha = alpha

        all_attrs = list(self.findStyledAttrs(style))
        skia_paint_kwargs = dict(AntiAlias=True)
        for attrs, attr in all_attrs:
            method, *args = attr
            if method == "skp":
                skia_paint_kwargs = args[0]
                if "AntiAlias" not in skia_paint_kwargs:
                    skia_paint_kwargs["AntiAlias"] = True
            elif method == "blendmode":
                self.blendmode = args[0].to_skia()

        for attrs, attr in all_attrs:
            filtered_paint_kwargs = {}
            for k, v in skia_paint_kwargs.items():
                if not k.startswith("_"):
                    filtered_paint_kwargs[k] = v
            self.paint = skia.Paint(**filtered_paint_kwargs)
            if self.blendmode:
                self.paint.setBlendMode(self.blendmode)
            method, *args = attr
            if method == "skp":
                pass
            elif method == "skb":
                pass
            elif method == "blendmode":
                pass
            elif method == "stroke" and args[0].get("weight") == 0:
                pass
            elif method == "dash":
                pass
            else:
                canvas.save()
                did_draw = self.applyDATAttribute(attrs, attr)
                self.paint.setAlphaf(self.paint.getAlphaf()*self.alpha)
                if not did_draw:
                    canvas.drawPath(self.path, self.paint)
                canvas.restore()
    
    def fill(self, color):
        self.paint.setStyle(skia.Paint.kFill_Style)
        if color:
            if isinstance(color, Gradient):
                self.gradient(color)
            elif isinstance(color, Color):
                self.paint.setColor(color.skia())
    
    def stroke(self, weight=1, color=None, dash=None):
        self.paint.setStyle(skia.Paint.kStroke_Style)
        if dash:
            self.paint.setPathEffect(skia.DashPathEffect.Make(*dash))
        if color and weight > 0:
            self.paint.setStrokeWidth(weight)
            if isinstance(color, Gradient):
                self.gradient(color)
            else:
                self.paint.setColor(color.skia())
    
    def gradient(self, gradient):
        self.paint.setShader(skia.GradientShader.MakeLinear([s[1].flip(self.rect).xy() for s in gradient.stops], [s[0].skia() for s in gradient.stops]))
    
    def image(self, src=None, opacity=1, rect=None, pattern=True):
        if isinstance(src, skia.Image):
            image = src
        else:
            image = skia.Image.MakeFromEncoded(skia.Data.MakeFromFileName(str(src)))

        if not image:
            print("image <", src, "> not found, cannot be used")
            return
        
        _, _, iw, ih = image.bounds()
        
        if pattern:
            matrix = skia.Matrix()
            matrix.setScale(rect.w / iw, rect.h / ih)
            self.paint.setShader(image.makeShader(
                skia.TileMode.kRepeat,
                skia.TileMode.kRepeat,
                matrix
            ))
        
        if opacity != 1:
            tf = skia.ColorFilters.Matrix([
                1, 0, 0, 0, 0,
                0, 1, 0, 0, 0,
                0, 0, 1, 0, 0,
                0, 0, 0, opacity, 0
            ])
            cf = self.paint.getColorFilter()
            if cf:
                self.paint.setColorFilter(skia.ColorFilters.Compose(
                    tf, cf))
            else:
                self.paint.setColorFilter(tf)
        
        if not pattern:
            bx, by, bw, bh = self.path.getBounds()
            if rect:
                bx, by = rect.flip(self.rect.h).xy()
                #bx += rx
                #by += ry
            sx = rect.w / iw
            sy = rect.h / ih
            self.canvas.save()
            #self.canvas.setMatrix(matrix)
            self.canvas.clipPath(self.path, doAntiAlias=True)
            if False:
                self.canvas.scale(sx, sy)
            else:
                # TODO scale the image, or maybe that shouldn't be here? this scaling method is horrible for image quality
                self.canvas.scale(sx, sy)
            was_alpha = self.paint.getAlphaf()
            self.paint.setAlphaf(was_alpha*self.alpha)
            self.canvas.drawImage(image, bx/sx, by/sy, self.paint)
            self.paint.setAlphaf(was_alpha)
            self.canvas.restore()
            return True
    
    def shadow(self, clip=None, radius=10, color=Color.from_rgb(0,0,0,1)):
        #print("SHADOW>", self.style, clip, radius, color)
        if clip:
            if isinstance(clip, Rect):
                skia.Rect()
                sr = skia.Rect(*clip.scale(self.scale, "mnx", "mny").flip(self.rect.h).mnmnmxmx())
                self.canvas.clipRect(sr)
            elif isinstance(clip, DATPen):
                sp = SkiaPathPen(clip, self.rect.h)
                self.canvas.clipPath(sp.path, doAntiAlias=True)
        self.paint.setColor(skia.ColorBLACK)
        self.paint.setImageFilter(skia.ImageFilters.DropShadow(0, 0, radius, radius, color.skia()))
        return
    
    def Composite(pens, rect, save_to, scale=1, context=None, style=None):
        if context:
            info = skia.ImageInfo.MakeN32Premul(rect.w, rect.h)
            surface = skia.Surface.MakeRenderTarget(context, skia.Budgeted.kNo, info)
        else:
            #print("CPU RENDER")
            surface = skia.Surface(rect.w, rect.h)
        
        with surface as canvas:
            if callable(pens):
                pens(canvas) # direct-draw
            else:
                SkiaPen.CompositeToCanvas(pens, rect, canvas, scale=scale, style=style)

        image = surface.makeImageSnapshot()
        image.save(save_to, skia.kPNG)
    
    def PDFOnePage(pens, rect, save_to, scale=1):
        stream = skia.FILEWStream(str(save_to))
        with skia.PDF.MakeDocument(stream) as document:
            with document.page(rect.w, rect.h) as canvas:
                SkiaPen.CompositeToCanvas(pens, rect, canvas, scale=scale)
    
    def PDFMultiPage(pages, rect, save_to, scale=1):
        stream = skia.FILEWStream(str(save_to))
        with skia.PDF.MakeDocument(stream) as document:
            for page in pages:
                with document.page(rect.w, rect.h) as canvas:
                    SkiaPen.CompositeToCanvas(page, rect, canvas, scale=scale)
    
    def SVG(pens, rect, save_to, scale=1):
        stream = skia.FILEWStream(str(save_to))
        canvas = skia.SVGCanvas.Make((rect.w, rect.h), stream)
        SkiaPen.CompositeToCanvas(pens, rect, canvas, scale=scale)
        del canvas
        stream.flush()
    
    def CompositeToCanvas(pens, rect, canvas, scale=1, style=None):
        if scale != 1:
            pens.scale(scale, scale, Point((0, 0)))
        
        if not pens.visible:
            return
        
        def draw(pen, state, data):
            if state != 0:
                return

            if not pen.visible:
                return
            
            if isinstance(pen, DATText):
                if not isinstance(pen.style, Style):
                    pen.style = Style(*pen.style[:-1], **pen.style[-1], load_font=0)
                
                if isinstance(pen.style.font, str):
                    font = skia.Typeface(pen.style.font)
                else:
                    font = skia.Typeface.MakeFromFile(str(pen.style.font.path))
                    if len(pen.style.variations) > 0:
                        fa = skia.FontArguments()
                        # h/t https://github.com/justvanrossum/drawbot-skia/blob/master/src/drawbot_skia/gstate.py
                        to_int = lambda s: struct.unpack(">i", bytes(s, "ascii"))[0]
                        makeCoord = skia.FontArguments.VariationPosition.Coordinate
                        rawCoords = [makeCoord(to_int(tag), value) for tag, value in pen.style.variations.items()]
                        coords = skia.FontArguments.VariationPosition.Coordinates(rawCoords)
                        fa.setVariationDesignPosition(skia.FontArguments.VariationPosition(coords))
                        font = font.makeClone(fa)
                pt = pen._frame.point("SW")
                canvas.drawString(
                    pen.text,
                    pt.x,
                    rect.h - pt.y,
                    skia.Font(font, pen.style.fontSize),
                    skia.Paint(AntiAlias=True, Color=pen.style.fill.skia()))
                return
            elif isinstance(pen, DATImage):
                paint = skia.Paint(AntiAlias=True)
                f = pen._frame
                canvas.save()
                for action, *args in pen.transforms:
                    if action == "rotate":
                        deg, pt = args
                        canvas.rotate(-deg, pt.x, pt.y)
                    #print(action, args)
                paint.setAlphaf(paint.getAlphaf()*data["alpha"])
                canvas.drawImage(pen._img, f.x, f.y, paint)
                canvas.restore()
                #pen = DATPen().rect(pen.bounds()).img(pen._img, rect=pen.bounds(), pattern=False)
                return
            
            if state == 0:
                SkiaPen(pen, rect, canvas, scale, style=style, alpha=data["alpha"])
        
        pens.walk(draw, visible_only=True)
    
    def Precompose(pens, rect, fmt=None, context=None, scale=1, disk=False):
        rect = rect.round()
        if context:
            info = skia.ImageInfo.MakeN32Premul(rect.w, rect.h)
            surface = skia.Surface.MakeRenderTarget(context, skia.Budgeted.kNo, info)
            assert surface is not None
        else:
            #print("CPU PRECOMPOSE")
            surface = skia.Surface(rect.w, rect.h)
        
        with surface as canvas:
            SkiaPen.CompositeToCanvas(pens.translate(-rect.x, -rect.y), rect, canvas)
        img = surface.makeImageSnapshot()
        if scale != 1:
            x, y, w, h = rect.scale(scale)
            img = img.resize(int(w), int(h))
        
        if disk:
            img.save(disk, skia.kPNG)
            #return disk
        return img
    
    def ReadImage(src):
        return skia.Image.MakeFromEncoded(skia.Data.MakeFromFileName(str(src)))


import re
import unicodedata
from itertools import groupby

def between(c, a, b):
    return ord(a) <= ord(c) <= ord(b)

LATIN = lambda c: between(c, '\u0000', '\u024F')
KATAKANA = lambda c: between(c, '\u30A0', '\u30FF')
HIRAGANA = lambda c: between(c, '\u3040', '\u309F')
CJK = lambda c: between(c, '\u4E00', '\u9FFF')

modes = [
    "LATIN",
    "ARABIC",
    "HEBREW",
    "SPACE",
    "CJK",
    "KATAKANA"
]

def segment(txt, mode="LATIN", includeNames=False):
    current_mode = mode
    runs = [[mode]]
    for i, c in enumerate(txt):
        try:
            n = unicodedata.name(c)
        except ValueError:
            n = "PRIVATE"
        for m in modes:
            if m in n:
                if current_mode != m:
                    current_mode = m
                    runs.append([m])
        runs[-1].append((n, c) if includeNames else c)
    if len(runs[0]) == 0:
        runs = runs[1:]
    
    for idx, (mode, *run) in enumerate(runs):
        p = runs[idx-1] if idx > 0 else None
        n = runs[idx+1] if len(runs) > idx+1 else None
        if p and n:
            if mode == "SPACE" and p[0] == n[0]:
                runs[idx][0] = p[0]
    
    grouped_runs = []
    for k, g in groupby(runs, lambda e: e[0]):
        txt = "".join(["".join(g[1:]) for g in list(g)])
        if txt:
            found_modes = set()
            for c in txt:
                try:
                    n = unicodedata.name(c)
                except ValueError:
                    n = "PRIVATE"
                for m in modes:
                    if m in n:
                        found_modes.add(m)
            grouped_runs.append((found_modes, txt))
    
    # reverse number ranges in arabic
    for idx, (cats, line) in enumerate(grouped_runs):
        if "ARABIC" in cats or "HEBREW" in cats:
            grouped_runs[idx] = (cats, re.sub("[0-9]+", lambda m: m.group()[::-1], line))

    return grouped_runs

if __name__ == "__main__":
    from pprint import pprint
    s = "ABC (جاف + رطب (ما قبل"
    s = "(رطب (ما قبل"
    #s = "Ali الملخبط Boba"
    s = "وصل الإستيرِوﬂLim/Satلل"
    runs = segment(s, "LATIN")
    pprint(runs)


# to be loaded from within Blender

from coldtype.geometry import Point, Line, Rect
from coldtype.pens.drawbotpen import DrawBotPen
from coldtype.pens.datpen import DATPen, DATPens
from coldtype.pens.blenderpen import BlenderPen, BPH
from coldtype.text.reader import StyledString, Style, Font
from coldtype.text.composer import StSt
from coldtype.color import hsl, bw
from pathlib import Path
try:
    import bpy
except ImportError:
    print("no bpy")
    pass

from coldtype.time.timeline import Timeline
from coldtype.time.timeable import Timeable
from coldtype.time import Frame

class b3d_animation(Timeable):
    def __init__(self, duration=10): # TODO possible to read from project?
        self.func = None
        self.name = None
        self.current_frame = -1
        self.start = 0
        self.end = duration
        self.timeline = Timeline(duration)
        self.t = self.timeline
    
    def update(self):
        self.func(Frame(self.current_frame, self))
    
    def __call__(self, func):
        self.func = func
        if not self.name:
            self.name = self.func.__name__

        def _frame_update_handler(scene):
            if scene.frame_current != self.current_frame:
                self.current_frame = scene.frame_current
                self.update()
        
        bpy.app.handlers.frame_change_post.clear() # TODO look for closure one only?
        bpy.app.handlers.frame_change_post.append(_frame_update_handler)

        self.current_frame = bpy.context.scene.frame_current
        self.update()
        return self


from ..misc.properties import cachedProperty
from ..misc.hbShape import characterGlyphMapping
from . import mergeScriptsAndLanguages


class BaseFont:

    def __init__(self, fontPath, fontNumber, dataProvider=None):
        self.fontPath = fontPath
        self.fontNumber = fontNumber
        self.cocoa = True
        self.resetCache()

    def resetCache(self):
        self._glyphDrawings = [{}, {}]  # cache for (outline, colorLayers) objects
        self._currentVarLocation = None  # used to determine whether to purge the outline cache
        # Invalidate cached properties
        del self.unitsPerEm
        del self.colorPalettes
        del self.featuresGSUB
        del self.featuresGPOS
        del self.stylisticSetNames
        del self.scripts
        del self.axes

    def close(self):
        pass

    def load(self, outputWriter):
        pass

    def updateFontPath(self, newFontPath):
        """This gets called when the source file was moved."""
        self.fontPath = newFontPath

    def getExternalFiles(self):
        """Subclasses may override this to return a list of external files,
        that clients can observe for changes.
        """
        return []

    def canReloadWithChange(self, externalFilePath):
        """ `externalFilePath` is None or an external file. If it is None,
        the main source file was changed on disk, else the externalFilePath
        was changed.

        This method should return True if it can update itself, in which case
        font.load() will be called. If it returns False, the font will be
        discarded and rebuilt from scratch.
        """
        return False

    @cachedProperty
    def unitsPerEm(self):
        return self.ttFont["head"].unitsPerEm

    @cachedProperty
    def colorPalettes(self):
        return None

    @cachedProperty
    def featuresGSUB(self):
        return set(self.shaper.getFeatures("GSUB"))

    @cachedProperty
    def featuresGPOS(self):
        return set(self.shaper.getFeatures("GPOS"))

    @cachedProperty
    def stylisticSetNames(self):
        return self.shaper.getStylisticSetNames()

    @cachedProperty
    def scripts(self):
        gsub = self.shaper.getScriptsAndLanguages("GSUB")
        gpos = self.shaper.getScriptsAndLanguages("GPOS")
        return mergeScriptsAndLanguages(gsub, gpos)

    @cachedProperty
    def axes(self):
        fvar = self.ttFont.get("fvar")
        if fvar is None:
            return {}
        name = self.ttFont["name"]
        axes = {}
        for axis in fvar.axes:
            axisDict = dict(name=str(name.getName(axis.axisNameID, 3, 1)),
                            minValue=axis.minValue,
                            defaultValue=axis.defaultValue,
                            maxValue=axis.maxValue)
            axes[axis.axisTag] = axisDict
        return axes

    def getGlyphRunFromTextInfo(self, textInfo, colorPalettesIndex=0, addDrawings=True, **kwargs):
        text = textInfo.text
        direction = textInfo.directionOverride
        script = textInfo.scriptOverride
        language = textInfo.languageOverride
        colorLayers = kwargs.get("colorLayers", False)
        kwargs.pop("colorLayers", None)

        if not self.colorPalettes:
            colorPalette = []
        else:
            colorPalette = self.colorPalettes[colorPalettesIndex]

        glyphs = GlyphsRun(len(text), self.unitsPerEm, direction in ("TTB", "BTT"), colorPalette)

        for segmentText, segmentScript, segmentBiDiLevel, firstCluster in textInfo.segments:
            if script is not None:
                segmentScript = script
            if direction is not None:
                segmentDirection = direction
            elif segmentBiDiLevel is None:
                segmentDirection = None  # Let HarfBuzz figure it out
            else:
                segmentDirection = ["LTR", "RTL"][segmentBiDiLevel % 2]
            run = self.getGlyphRun(segmentText,
                                   direction=segmentDirection,
                                   script=segmentScript,
                                   language=language,
                                   **kwargs)
            for gi in run:
                gi.cluster += firstCluster
            glyphs.extend(run)

        x = y = 0
        for gi in glyphs:
            gi.pos = x + gi.dx, y + gi.dy
            x += gi.ax
            y += gi.ay
        glyphs.endPos = (x, y)

        if addDrawings:
            self.addGlyphDrawings(run, colorLayers=colorLayers)
        return glyphs

    def getGlyphRun(self, text, *, features=None, varLocation=None,
                    direction=None, language=None, script=None):
        self.setVarLocation(varLocation)
        glyphInfo = self.shaper.shape(text, features=features, varLocation=varLocation,
                                      direction=direction, language=language, script=script)
        return glyphInfo

    def setVarLocation(self, varLocation):
        axes = self.axes
        if varLocation:
            # subset to our own axes
            varLocation = {k: v for k, v in varLocation.items() if k in axes}
        if self._currentVarLocation != varLocation:
            self._purgeCaches()
            self._currentVarLocation = varLocation
            self.varLocationChanged(varLocation)
    
    def addGlyphDrawings(self, glyphs, colorLayers=False):
        glyphNames = (gi.name for gi in glyphs)
        for glyph, glyphDrawing in zip(glyphs, self.getGlyphDrawings(glyphNames, colorLayers)):
            glyph.glyphDrawing = glyphDrawing

    def getGlyphDrawings(self, glyphNames, colorLayers=False):
        for glyphName in glyphNames:
            glyphDrawing = self._glyphDrawings[colorLayers].get(glyphName)
            if glyphDrawing is None:
                glyphDrawing = self._getGlyphDrawing(glyphName, colorLayers)
                self._glyphDrawings[colorLayers][glyphName] = glyphDrawing
            yield glyphDrawing

    def _purgeCaches(self):
        self._glyphDrawings = [{}, {}]

    def _getGlyphDrawing(self, glyphName, colorLayers):
        raise NotImplementedError()

    def varLocationChanged(self, varLocation):
        # Optional override
        pass


class GlyphsRun(list):

    def __init__(self, numChars, unitsPerEm, vertical, colorPalette=None):
        self.numChars = numChars
        self.unitsPerEm = unitsPerEm
        self.vertical = vertical
        self._glyphToChars = None
        self._charToGlyphs = None
        self.endPos = (0, 0)
        self.colorPalette = [] if colorPalette is None else colorPalette

    def mapGlyphsToChars(self, glyphIndices):
        if self._glyphToChars is None:
            self._calcMappings()
        glyphToChars = self._glyphToChars
        return {ci for gi in glyphIndices for ci in glyphToChars[gi]}

    def mapCharsToGlyphs(self, charIndices):
        if self._charToGlyphs is None:
            self._calcMappings()
        charToGlyphs = self._charToGlyphs
        return {gi for ci in charIndices for gi in charToGlyphs[ci]}

    def _calcMappings(self):
        clusters = [glyphInfo.cluster for glyphInfo in self]
        self._glyphToChars, self._charToGlyphs = characterGlyphMapping(clusters, self.numChars)

    @cachedProperty
    def glyphNames(self):
        return [glyphInfo.name for glyphInfo in self]



#from coldtype.pens.draftingpens import DraftingPen, DraftingPens
from coldtype.pens.draftingpens import DraftingPens
from coldtype.geometry import Rect, Point

from coldtype.text.shaper import segment
from coldtype.text.reader import Style, StyledString, FittableMixin, Font, normalize_font_path, SegmentedString,  _PenClass, _PensClass


class GrafStyle():
    def __init__(self, leading=10, x="centerx", xp=0, width=0, **kwargs):
        self.leading = kwargs.get("l", leading)
        self.x = x
        self.xp = xp
        self.width = width


class Graf():
    def __init__(self, lines, container, style=None, **kwargs):
        if isinstance(container, Rect):
            self.container = _PenClass().rect(container)
        else:
            self.container = container
        if style and isinstance(style, GrafStyle):
            self.style = style
        else:
            self.style = GrafStyle(**kwargs)
        self.lines = lines
    
    def lineRects(self):
        # which came first, the height or the width???
        rects = []
        leadings = []
        box = self.container.ambit()
        leftover = box
        for l in self.lines:
            box, leftover = leftover.divide(l.height(), "maxy", forcePixel=True)
            if self.style.leading < 0:
                # need to add pixels back to leftover
                leftover.h += abs(self.style.leading)
            else:
                leading, leftover = leftover.divide(self.style.leading, "maxy", forcePixel=True)
                leadings.append(leading)
            rects.append(box)
        return rects
    
    def width(self):
        if self.style.width > 0:
            return self.style.width
        else:
            return max([l.width() for l in self.lines])

    def fit(self, width=None):
        rects = self.lineRects()
        for idx, l in enumerate(self.lines):
            if width:
                fw = width
            else:
                fw = rects[idx].w - self.style.xp
            l.fit(fw)
        return self
    
    def pens(self, align=False):
        rects = self.lineRects()
        pens = _PensClass()
        for idx, l in enumerate(self.lines):
            r = rects[idx]
            dps = l.pens().translate(r.x, r.y) # r.x
            dps.addFrame(Rect(r.x, r.y, r.w, r.h))
            #if align:
            #    dps.container = Rect(0, r.y, r.w, r.h)
            #    dps.align(dps.container, x=self.style.x, y=None)
            pens.append(dps)
        return pens
    
    def fpa(self, rect=None):
        return self.fit().pens().align(rect or self.container.ambit())


class Lockup(FittableMixin):
    def __init__(self, slugs, preserveLetters=True, nestSlugs=True):
        self.slugs = slugs
        self.preserveLetters = preserveLetters
        self.nestSlugs = nestSlugs
    
    def width(self):
        return sum([s.width() for s in self.slugs])
    
    def height(self):
        return max([s.height() for s in self.slugs])
    
    def textContent(self):
        return "/".join([s.textContent() for s in self.slugs])

    def shrink(self):
        adjusted = False
        for s in self.slugs:
            adjusted = s.shrink() or adjusted
        return adjusted

    def pens(self):
        pens = []
        x_off = 0
        for s in self.slugs:
            try:
                x_off += s.margin[0]
            except:
                pass
            if self.preserveLetters:
                dps = s.pens()
                dps.translate(x_off, 0)
                if self.nestSlugs:
                    pens.append(dps)
                else:
                    pens.extend(dps._pens)
            else:
                dps = s.pen()
                dps.translate(x_off, 0)
                pens.append(dps)
            x_off += dps.ambit().w
            try:
                x_off += s.margin[1]
                x_off += s.strings[-1].tracking
            except:
                pass
        return _PensClass(pens)
    
    def pen(self):
        return self.pens().pen()
    
    def TextToLines(text, primary, fallback=None):
        lines = []
        for line in text.split("\n"):
            lines.append(Lockup([Slug(line, primary, fallback)]))
        return lines
    
    def SlugsToLines(slugs):
        return [Lockup([slug]) for slug in slugs]


def T2L(text, primary, fallback=None):
    return Lockup.TextToLines(text, primary, fallback)


class Slug(SegmentedString):
    def __init__(self, text, primary, fallback=None):
        self.text = text
        self.primary = primary
        self.fallback = fallback
        self.strings = []
        self.tag()
    
    def tag(self):
        if self.fallback:
            segments = segment(self.text, "LATIN")
            self.strings = [StyledString(s[1], self.fallback if "LATIN" in s[0] else self.primary) for s in segments]
        else:
            self.strings = [StyledString(self.text, self.primary)]
    
    def pen(self):
        return self.pens().pen()
    
    def LineSlugs(text, primary, fallback=None):
        lines = []
        for line in text.split("\n"):
            lines.append(Slug(line, primary, fallback))
        return lines


class Composer():
    """
    For multiline text lockups
    """
    def __init__(self, rect:Rect, text:str, style:Style, leading=10, fit=None):
        lockups = Slug.LineSlugs(text, style)
        self.rect = rect
        self.graf = Graf(lockups, self.rect, leading=leading)
        if fit is not None:
            self.graf.fit(fit)
    
    def pens(self):
        """
        Structured representation of the multi-line text
        
        In the return ``DATPens``, each line will be a ``DATPens``, then within those lines, each glyph/ligature for that line will be an individual ``DATPen``
        """
        return self.graf.pens()
    
    def pen(self):
        """
        Entire multiline text as a single vector
        """
        return self.graf.pens().pen()


def StSt(text,
    font,
    font_size=24,
    rect=Rect(1080, 1080),
    xa="mdx",
    **kwargs):
    
    if isinstance(font, Style):
        style = font
    else:
        style = Style(font, font_size, **kwargs)
    
    fit = kwargs.get("fit", None)
    leading = kwargs.get("leading", 10)

    if "\n" in text:
        lockup = Composer(rect, text, style, fit=fit, leading=leading).pens()
        if xa:
            lockup = lockup.xa(xa)
    else:
        lockup = StyledString(text, style)
        if fit:
            lockup.fit(fit)
        lockup = lockup.pens()
    return lockup

def Glyphwise(st, styler):
    def kx(dps, idx):
        return dps[idx].ambit().x
    
    def krn(off, on, idx):
        return kx(off, idx) - kx(on, idx)

    dps = DraftingPens()
    prev = 0
    tracks = []

    for idx, c in enumerate(st):
        test = c
        target = 0
        if idx < len(st) - 1:
            test = test + st[idx+1]
        if idx > 0:
            test = st[idx-1] + test
            target = 1
        
        skon = styler(idx, c)
        skoff = skon.mod(kern=0)

        tkon = StSt(test, skon)
        tkoff = StSt(test, skoff)

        if idx == 0:
            dps.append(tkoff[0])
            prev = krn(tkoff, tkon, 1)
        if target > 0:
            _prev = krn(tkoff, tkon, 1)
            prev_av = (prev+_prev)/2
            #print(prev_av, kx(tkoff, 1))

            dps.append(tkoff[1].copy(with_data=True)
                .translate(-kx(tkoff, 1), 0))
            tracks.append(-prev_av)

            if len(test) > 2:
                #print("_PREV", _prev)
                _next = krn(tkoff, tkon, 2) - _prev #tkoff[0].ambit().w
                #print("next", _next)
            else:
                _next = 0
            prev = _next
    
    #print(tracks)
    return dps.distribute(
        tracks=tracks
    )


import math
from fontTools.misc.transform import Transform
from coldtype.geometry.geometrical import Geometrical
from coldtype.geometry.point import Point
from coldtype.geometry.primitives import line_intersection, calc_angle, polar_coord
from coldtype.interpolation import norm


class Line(Geometrical):
    def __init__(self, start, end):
        self.start = Point(start)
        self.end = Point(end)
    
    def __eq__(self, l):
        if not hasattr(l, "start"):
            return False
        return self.start == l.start and self.end == l.end

    __hash__ = object.__hash__

    def point(self, p):
        if p == "N":
            return self.pn
        elif p == "E":
            return self.pe
        elif p == "S":
            return self.ps
        elif p == "W":
            return self.pw
        elif p == "C":
            return self.mid
    
    @property
    def mid(self):
        return self.start.i(0.5, self.end)
    
    @property
    def mxx(self):
        return max([p.x for p in self.pts()])
    
    @property
    def mnx(self):
        return min([p.x for p in self.pts()])
    
    @property
    def mxy(self):
        return max([p.y for p in self.pts()])
    
    @property
    def mny(self):
        return min([p.y for p in self.pts()])
    
    @property
    def pe(self):
        return max(self.pts(), key=lambda p: p.x)
    
    @property
    def pw(self):
        return min(self.pts(), key=lambda p: p.x)
    
    @property
    def pn(self):
        return max(self.pts(), key=lambda p: p.y)
    
    @property
    def ps(self):
        return min(self.pts(), key=lambda p: p.y)
    
    def __repr__(self):
        return f"Line({self.start}, {self.end})"
    
    def __len__(self):
        return 2
    
    def __getitem__(self, idx):
        if idx == 0:
            return self.start
        elif idx == 1:
            return self.end
        else:
            raise IndexError("Line only has two points")

    def reverse(self):
        p1, p2 = self
        return Line(p2, p1)
    
    def __invert__(self):
        return self.reverse()

    def t(self, t):
        return self.start.interp(t, self.end)
    
    def length(self):
        a2 = math.pow(self.start.x - self.end.x, 2)
        b2 = math.pow(self.start.y - self.end.y, 2)
        return math.sqrt(a2 + b2)
    
    @property
    def l(self):
        return self.length()
    
    def tpx(self, tpx, limit=True):
        x = tpx * math.cos(self.angle())
        y = tpx * math.sin(self.angle())
        tp = self.start.offset(x, y)
        if not limit:
            return tp
        else:
            if Line(self.start, tp).length() > self.length():
                return self.end
            else:
                return tp

    def angle(self):
        return calc_angle(self.start, self.end)
    
    @property
    def ang(self):
        return self.angle()%math.pi
    
    def pts(self):
        return [self.start, self.end]
    
    def transform(self, t):
        pts = self.pts()
        x1, x2 = [t.transformPoint(pt) for pt in pts]
        return Line(x1, x2)

    def rotate(self, degrees, point=None):
        if Transform:
            t = Transform()
            if not point:
                point = self.mid
            t = t.translate(point.x, point.y)
            t = t.rotate(math.radians(degrees))
            t = t.translate(-point.x, -point.y)
            return self.transform(t)
        else:
            raise Exception("fontTools not installed")
    
    def bow(self, amt, t=0.5, angle=90):
        rotated = self.rotate(angle, point=self.t(t))
        return rotated.tpx(self.length()*0.5+amt)
    
    def project(self, pt, dist, angle=90):
        dx, dy = polar_coord((0, 0), self.ang+math.radians(angle), dist)
        return self.t(pt).offset(dx, dy)
    
    def inset(self, px):
        return Line(self.tpx(px), self.reverse().tpx(px))
    
    inset_y = inset
    inset_x = inset
    
    def extr(self, amt):
        p1, p2 = self
        return Line(p2.i(1+amt, p1), p1.i(1+amt, p2))
    
    def offset(self, x, y):
        p1, p2 = self
        return Line(p1.offset(x, y), p2.offset(x, y))
    
    def offset_x(self, dx):
        return self.offset(dx, 0)
    
    def offset_y(self, dy):
        return self.offset(0, dy)
    
    def tan_out(self):
        return self
    
    def tan_in(self):
        return self
    
    o = offset
    
    def __floordiv__(self, other):
        return self.offset(0, other)
    
    def __truediv__(self, other):
        return self.offset(other, 0)
    
    def __mod__(self, other):
        return self.offset(*other)
    
    def intersection(self, other):
        return Point(line_intersection(self, other))
    
    def __and__(self, other):
        return self.intersection(other)
    
    def join(self, other):
        from coldtype.geometry.rect import Rect
        return Rect.FromPoints(self.start, self.end, other.end, other.start)
    
    def interp(self, x, other):
        return Line(self.start.i(x, other.start), self.end.i(x, other.end))

    i = interp

    def setx(self, x):
        return Line(self.start.setx(x), self.end.setx(x))
    
    def __mul__(self, other):
        return self.setx(other)
    
    def sety(self, y):
        return Line(self.start.sety(y), self.end.sety(y))
    
    def __matmul__(self, other):
        return self.sety(other)


from watchdog.observers import Observer


class AsyncEventHandler(object):
    def __init__(self, loop=None, on_modified=None):
        self.on_modified = on_modified

    def dispatch(self, event):
        method_name = f"on_{event.event_type}"
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            method(event)


class AsyncWatchdog(object):
    def __init__(self, path, recursive=True, on_modified=None):
        self._observer = Observer()
        evh = AsyncEventHandler(on_modified=on_modified)
        self._observer.schedule(evh, str(path), recursive)

    def start(self):
        self._observer.start()

    def stop(self):
        self._observer.stop()
        self._observer.join()


import datetime

from subprocess import run
from defcon import Font as DFont
from coldtype.geometry.rect import Rect
from coldtype.helpers import glyph_to_uni
from coldtype.time.timeline import Timeline
from coldtype.renderable import renderable, animation
from coldtype.text.reader import Style, StyledString, Font
from coldtype.pens.datpen import DATPen, DATPens
from coldtype.pens.dattext import DATText
from coldtype.color import hsl
from pathlib import Path

from typing import Tuple


class glyphfn():
    def __init__(self, width=1000, lsb=0, rsb=0):
        """lsb = left-side-bearing / rsb = right-side-bearing"""
        self.width = width
        self.lsb = lsb
        self.rsb = rsb
        self.frame = None
        self.bbox = None
    
    def add_font(self, font):
        self.frame = Rect(self.width, font.ufo.info.capHeight)
        self.bbox = Rect(self.lsb + self.width + self.rsb, font.ufo.info.capHeight)
        return self
    
    def __call__(self, func):
        self.func = func
        self.glyph_name = func.__name__
        self.unicode = glyph_to_uni(self.glyph_name)
        return self


class generativefont(animation):
    def __init__(self,
        lookup,
        ufo_path:Path,
        font_name="Test",
        style_name="Regular",
        cap_height=750,
        ascender=750,
        descender=-250,
        units_per_em=1000,
        preview_size=(1000, None)):

        pw, ph = preview_size
        self.preview_frame = Rect(pw, ph if ph else (-descender*2) + cap_height)

        if not ufo_path.exists():
            ufo = DFont()
            ufo.save(ufo_path)
        else:
            ufo = DFont(str(ufo_path))
        
        ufo.info.familyName = font_name
        ufo.info.styleName = style_name
        ufo.info.capHeight = cap_height
        ufo.info.ascender = ascender
        ufo.info.descender = descender
        ufo.info.unitsPerEm = units_per_em
        
        self.ufo = ufo

        super().__init__(
            self.preview_frame, timeline=self.timeline(lookup),
            postfn=generativefont.ShowGrid)
    
    def _find_glyph_fns(self, lookup):
        """
        `lookup` should probably be `globals()`
        """
        self.glyph_fns = []
        itms = lookup.items()
        for k, v in itms:
            if isinstance(v, glyphfn):
                self.glyph_fns.append(v)
    
    def timeline(self, lookup):
        self._find_glyph_fns(lookup)
        return Timeline(len(self.glyph_fns))
    
    def frame_to_fn(self, fi) -> Tuple[str, dict]:
        return [
            f"def {self.glyph_fns[fi].glyph_name}(",
            dict(decorator="@glyphfn")]
    
    def ShowGrid(render, result):
        if False: # flip to true if you don't want to see the grid
            return result

        gfn = result[0].data.get("gfn")
        if not gfn:
            print("! No glyph found")
        else:
            bbox = gfn.bbox.offset(0, 250)
            return DATPens([
                DATPens(result).translate(0, 250),
                DATPen().gridlines(render.rect).s(hsl(0.6, a=0.3)).sw(1).f(None),
                (DATPen()
                    .line(bbox.es.extr(-100))
                    .line(bbox.en.extr(-100))
                    .line(bbox.ee.extr(-100))
                    .f(None).s(hsl(0.9, 1, a=0.5)).sw(4)),
                (DATText(gfn.glyph_name, Style("Times", 48, load_font=0),
                    render.rect.inset(50)))])
    
    def glyph_viewer(self, f):
        glyph_fn = self.glyph_fns[f.i]
        glyph_fn.add_font(self)

        print(f"> drawing :{glyph_fn.glyph_name}:")
        glyph_pen = glyph_fn.func(glyph_fn.frame).f(0)

        # shift over by the left-side-bearing
        glyph_pen.translate(glyph_fn.lsb, 0)
        glyph = glyph_pen.to_glyph(
            name=glyph_fn.glyph_name,
            width=glyph_fn.frame.w + glyph_fn.lsb + glyph_fn.rsb,
            allow_blank = True)
        glyph.unicode = glyph_to_uni(glyph_fn.glyph_name)
        self.ufo.insertGlyph(glyph)
        self.ufo.save()
        return glyph_pen.add_data("gfn", glyph_fn)
    
    def spacecenter(self, r, text):
        """
        This function loads the ufo that’s been created by the code above and displays it "as a font" (i.e. it compiles the ufo to a font and then uses the actual font to do standard font-display logic)
        """
        ufo = Font(self.ufo.path)
        return (StyledString("ABC CBA",
            Style(ufo, 150))
            .pens()
            .align(r)
            .f(0))

    def fontmake(self):
        ufo = DFont(self.ufo.path)
        date = datetime.datetime.now().strftime("%y%m%d%H%M")
        font_name = "_".join([
            ufo.info.familyName.replace(" ", ""),
            ufo.info.styleName.replace(" ", ""),
            date
        ])
        fontmade_path = self.ufo.path.parent / f"fontmakes/{font_name}.otf"
        fontmade_path.parent.mkdir(exist_ok=True)
        run([
            "fontmake",
            "-u", str(self.ufo.path),
            "-o", "otf",
            "--output-path=" + str(fontmade_path)])


from fontTools.misc.arrayTools import unionRect
if False:
    try:
        from ..mac.drawing import rectFromNSRect
    except:
        pass
from ..misc.properties import cachedProperty


class GlyphDrawing:

    def __init__(self, layers=None):
        self.layers = layers

    def appendPath(self, path, colorID=None):
        self.layers.append((path, colorID))

    @cachedProperty
    def bounds(self):
        bounds = None
        for path, colorID in self.layers:
            if not path.elementCount():
                continue
            pathBounds = rectFromNSRect(path.controlPointBounds())
            if bounds is None:
                bounds = pathBounds
            else:
                bounds = unionRect(bounds, pathBounds)
        return bounds

    def draw(self, colorPalette, defaultColor):
        for path, colorID in self.layers:
            color = colorPalette.get(colorID, defaultColor)
            color.set()
            path.fill()

    def pointInside(self, pt):
        return any(path.containsPoint_(pt) for path, colorID in self.layers)



from pathlib import Path
from collections import OrderedDict
from functools import partial

import unicodedata, math

from fontTools.misc.transform import Transform
from fontTools.pens.transformPen import TransformPen
from fontTools.pens.svgPathPen import SVGPathPen
from fontTools.pens.recordingPen import RecordingPen, replayRecording
from fontTools.pens.boundsPen import ControlBoundsPen, BoundsPen
from fontTools.ttLib.ttFont import TTFont

from coldtype.color import normalize_color
from coldtype.pens.draftingpen import DraftingPen
from coldtype.pens.draftingpens import DraftingPens
from coldtype.pens.datpen import DATPen, DATPens
from coldtype.geometry import Rect, Point

from typing import Optional, Callable, Union

_PenClass = DATPen
_PensClass = DATPens

# try:
#     from fontgoggles.font import getOpener
#     from fontgoggles.font.baseFont import BaseFont
#     from fontgoggles.font.otfFont import OTFFont
#     from fontgoggles.misc.textInfo import TextInfo
# except ModuleNotFoundError:

from coldtype.fontgoggles.font import getOpener
from coldtype.fontgoggles.font.baseFont import BaseFont
from coldtype.fontgoggles.font.otfFont import OTFFont
from coldtype.fontgoggles.misc.textInfo import TextInfo


class FittableMixin():
    def textContent(self):
        print("textContent() not overwritten")

    def fit(self, width):
        """Use various methods (tracking, `wdth` axis, etc. — properties specified in the `Style` object) to fit a piece of text horizontally to a given `width` (warning: not very fast)"""
        if isinstance(width, Rect):
            width = width.w
        current_width = self.width()
        tries = 0
        if current_width > width: # need to shrink
            while tries < 100000 and current_width > width:
                adjusted = self.shrink()
                if adjusted:
                    tries += 1
                    current_width = self.width()
                else:
                    #print(">>> TOO BIG :::", self.textContent())
                    return self
        elif current_width < width: # need to expand
            pass
        #print(">>>>>>>>>>>>>>>>>> FINAL TRIES", tries)
        return self


_prefixes = [
    ["¬", "~/Library/Fonts"],
    ["", "/Library/Fonts"]
]

class FontNotFoundException(Exception):
    pass

def normalize_font_prefix(path_string):
    for prefix, expansion in _prefixes:
        path_string = path_string.replace(prefix, expansion)
    return Path(path_string).expanduser().resolve()

def normalize_font_path(font, nonexist_ok=False):
    global _prefixes
    literal = normalize_font_prefix(str(font))
    ufo = literal.suffix == ".ufo"
    if nonexist_ok:
        return str(literal)
    if literal.exists() and (not literal.is_dir() or ufo):
        return str(literal)
    else:
        raise FontNotFoundException(literal)

FontCache = {}

class Font():
    # TODO support glyphs?
    def __init__(self, path, number=0, cacheable=False):
        self.path = Path(normalize_font_path(path))
        numFonts, opener, getSortInfo = getOpener(self.path)
        self.font:BaseFont = opener(self.path, number)
        self.font.cocoa = False
        self.cacheable = cacheable
        self._loaded = False
        self.load()
    
    def load(self):
        if self._loaded:
            return self
        else:
            self.font.load(None)
            self._loaded = True
            return self
    
    def Cacheable(path):
        if path not in FontCache:
            FontCache[path] = Font(path, cacheable=True).load()
        return FontCache[path]

class Style():
    """
    Class for configuring font properties

    **Keyword arguments**

    * ``font``: can either be a ``coldtype.text.Font`` object, a ``pathlib.Path``, or a plain string path
    * ``fontSize``: standard point-based font-size, expressed as integer
    * ``tracking`` (aka ``tu``): set the tracking, by default **in font-source-point-size** aka as if the font-size was always 1000; this means tracking is by default done relatively rather than absolutely (aka the relative tracking will not change when you change the fontSize)
    * ``trackingMode``: set to 0 to set tracking in a classic fontSize-based (defaults to 1, as described just above)
    * ``space``: set this to override the width of the standard space character (useful when setting text on a curve and the space is getting collapsed)
    * ``baselineShift`` (aka ``bs``): if an integer, shifts glyphs by that amount in y axis; if a list, shifts glyphs at corresponding index in list by that amount in y axis
    * ``xShift`` (aka ``xs``): if an integer, shifts glyphs by that amount in x axis; if a list, shifts glyphs at corresponding index in list by that amount in x axis
    * ``rotate``: rotate glyphs by degree
    * ``reverse`` (aka ``r``): reverse the order of the glyphs, so that the left-most glyph is first in when vectorized via ``.pens()``
    * ``removeOverlaps`` (aka ``ro``): automatically use skia-pathops to remove overlaps from the glyphs (useful when using variable ttf fonts)
    * ``lang``: set language directly, to access language-specific alternate characters/rules

    **Shorthand kwargs**

    * ``kp`` for ``kern_pairs`` — a dict of glyphName->[left,right] values in font-space
    * ``tl`` for ``trackingLimit``
    * ``bs`` for ``baselineShift``
    * ``ch`` for ``capHeight`` — a number in font-space; not specified, read from font; specified as 'x', capHeight is set to xHeight as read from font
    """
    def RegisterShorthandPrefix(prefix, expansion):
        global _prefixes
        _prefixes.append([prefix, str(expansion)])

    def __init__(self,
            font:Union[Font, str]=None,
            fontSize:int=12,            
            tracking=0,
            trackingMode=1,
            kern_pairs=dict(),
            space=None,
            baselineShift=0,
            xShift=None,
            rotate=0,
            reverse=False,
            removeOverlap=False,
            lang=None,
            narrower=None,
            include_blanks=False,
            palette=0,
            capHeight=None,
            data={},
            layer=None,
            liga=True,
            kern=True,
            fill=(0, 0.5, 1),
            stroke=None,
            strokeWidth=0,
            variations=dict(),
            variationLimits=dict(),
            trackingLimit=0,
            scaleVariations=True,
            mods=None,
            features=dict(),
            increments=dict(),
            varyFontSize=False,
            preventHwid=False,
            fitHeight=None,
            load_font=True, # should we attempt to load the font?
            tag=None, # way to differentiate in __eq__
            **kwargs):

        self.input = locals()
        self.input["self"] = None

        if load_font:
            if isinstance(font, Path):
                font = str(font)
            if isinstance(font, str):
                self.font:Font = Font(font)
                self.font.load() #font._syncLoad(None)
            else:
                self.font:Font = font
        else:
            self.font = font

        self.narrower = narrower
        self.layer = layer
        self.reverse = kwargs.get("r", reverse)
        self.removeOverlap = kwargs.get("ro", removeOverlap)
        self.rotate = rotate
        self.include_blanks = include_blanks
        self.scaleVariations = scaleVariations
        self.tag = tag

        try:        
            if "OS/2" in self.font.font.ttFont:
                os2 = self.font.font.ttFont["OS/2"]
                self.capHeight = os2.sCapHeight if hasattr(os2, "sCapHeight") else 0
                if self.capHeight == 0:
                    self.capHeight = os2.sTypoAscender if hasattr(os2, "sTypoAscender") else 750
            elif hasattr(self.font.font, "info"):
                self.capHeight = self.font.font.info.capHeight
            elif hasattr(self.font.font, "defaultInfo"):
                self.capHeight = self.font.font.defaultInfo.capHeight

            if capHeight: # override whatever the font says
                if capHeight != "x":
                    self.capHeight = capHeight
        except AttributeError:
            pass

        if fitHeight:
            self.fontSize = (fitHeight/self.capHeight)*1000
        else:
            self.fontSize = fontSize

        self.tracking = kwargs.get("t", tracking)
        self.kern_pairs = kwargs.get("kp", kern_pairs)
        self.trackingMode = trackingMode
        self.trackingLimit = kwargs.get("tl", trackingLimit)
        self.baselineShift = kwargs.get("bs", baselineShift)
        self.increments = increments
        self.space = space
        self.xShift = kwargs.get("xs", xShift)
        self.mods = mods
        self.palette = palette
        self.lang = lang
        self.data = data
        self.preventHwid = preventHwid

        if kwargs.get("tu"):
            self.trackingMode = 1 # this is the default now
            self.tracking = kwargs.get("tu")
            if not self.increments.get("tracking"):
                self.increments["tracking"] = 5 # TODO good?

        found_features = features.copy()
        for k, v in kwargs.items():
            if k.startswith("ss") and len(k) == 4:
                found_features[k] = v
            if k in ["dlig", "swsh", "onum", "tnum", "palt", "salt"]:
                found_features[k] = v
            if k in ["slig"]:
                if k == 0:
                    found_features[k] = 0
        
        self.features = {**dict(kern=kern, liga=liga), **found_features}

        self.fill = normalize_color(fill)
        self.stroke = normalize_color(stroke)
        if stroke and strokeWidth == 0:
            self.strokeWidth = 1
        else:
            self.strokeWidth = strokeWidth

        unnormalized_variations = variations.copy()
        self.axes = OrderedDict()
        self.variations = dict()
        self.variationLimits = dict()
        self.varyFontSize = varyFontSize
        
        if load_font and self.font.font.ttFont:
            try:
                fvar = self.font.font.ttFont['fvar']
            except:
                fvar = None
            if fvar:
                for axis in fvar.axes:
                    self.axes[axis.axisTag] = axis
                    self.variations[axis.axisTag] = axis.defaultValue
                    if axis.axisTag == "wdth": # the only reasonable default
                        self.variationLimits[axis.axisTag] = axis.minValue
                    if axis.axisTag in kwargs and axis.axisTag not in variations:
                        unnormalized_variations[axis.axisTag] = kwargs[axis.axisTag]

            self.addVariations(unnormalized_variations)
    
    def __eq__(self, other):
        if not self.tag == other.tag:
            return False
        if not self.font == other.font:
            #print("different font")
            return False
        elif not self.fontSize == other.fontSize:
            #print("different fontSize")
            return False
        
        for key, value in self.variations.items():
            if self.variations[key] != other.variations[key]:
                #print("different variation")
                return False
            
        if self.rotate != other.rotate:
            #print("different rotate")
            return False
        
        if self.fill != other.fill:
            #print("different fill")
            return False
        
        return True

    def mod(self, **kwargs):
        """Modify this style object to create a new
        one; kwargs can have all of the same kwargs as
        the standard `Style` constructor"""
        keyed = dict(**self.input, **self.input["kwargs"])
        del keyed["kwargs"]
        del keyed["self"]
        keyed.update(kwargs)
        return Style(**keyed)
    
    def addVariations(self, variations, limits=dict()):
        for k, v in self.normalizeVariations(variations).items():
            self.variations[k] = v
        for k, v in self.normalizeVariations(limits).items():
            self.variationLimits[k] = v
    
    def normalizeVariations(self, variations):
        scale = self.scaleVariations
        for k, v in variations.items():
            try:
                axis = self.axes[k]
            except KeyError:
                #print("Invalid axis", self.fontFile, k)
                continue
                #raise Exception("Invalid axis", self.fontFile, k)
            if v == "min":
                variations[k] = axis.minValue
            elif v == "max":
                variations[k] = axis.maxValue
            elif v == "default":
                variations[k] = axis.defaultValue
            elif scale:
                _v = max(0, min(1, v))
                variations[k] = float(abs(axis.maxValue-axis.minValue)*_v + axis.minValue)
            else:
                if v < axis.minValue or v > axis.maxValue:
                    variations[k] = max(axis.minValue, min(axis.maxValue, v))
                    print("----------------------")
                    print("Invalid Font Variation")
                    print(self.fontFile, self.axes[k].axisTag, v)
                    print("> setting", v, "<to>", variations[k])
                    print("----------------------")
                else:
                    variations[k] = v
        return variations
    
    def StretchX(flatten=10, debug=0, **kwargs):
        d = {}
        
        def stretcher(w, xp, i, p):
            np = (p.flatten(flatten) if flatten else p).nonlinear_transform(lambda x,y: (x if x < xp else x + w, y))
            if debug:
                (np.record(_PenClass()
                    .line([(xp, -250), (xp, 1000)])
                    .outline()))
            return np

        def is_left(a, b, c):
            return ((b[0] - a[0])*(c[1] - a[1]) - (b[1] - a[1])*(c[0] - a[0])) > 0
        
        def stretcher_slnt(w, xy, angle, i, p):
            if abs(angle) in [90, 270]:
                return p
            x0, y0 = xy
            ra = math.radians(90+angle)
            xdsc = x0 + (-250 - y0) / math.tan(ra)
            xasc = x0 + (1000 - y0) / math.tan(ra)

            np = (p.flatten(flatten) if flatten else p).nonlinear_transform(lambda x,y: (x if is_left((xdsc, -250), (xasc, 1000), (x, y)) else x + w, y))
            if debug:
                (np
                    .record(_PenClass()
                        .line([(xdsc, -250), (xasc, 1000)])
                        .outline())
                    .record(_PenClass()
                        .moveTo((x0+50/2, y0+50/2))
                        .dots(radius=50)))
            return np
        
        for k, v in kwargs.items():
            if len(v) == 3:
                d[k] = (v[0], partial(stretcher_slnt, v[0], v[1], v[2]))
            elif len(v) == 2:
                d[k] = (v[0], partial(stretcher, v[0], v[1]))
            elif len(v) == 1:
                pass
        return d
    
    def StretchY(flatten=10, align="mdy", debug=0, **kwargs):
        d = {}
        
        def stretcher(h, yp, i, p):
            np = (p.flatten(flatten) if flatten else p).nonlinear_transform(lambda x,y: (x, y if y < yp else y + h))
            if align == "mdy":
                np.translate(0, -h/2)
            elif align == "mxy":
                np.translate(0, -h)
            if debug:
                (np.record(_PenClass()
                    .line([(0, yp), (p.ambit().point("E").x, yp)])
                    .outline()))
            return np

        def is_left(a, b, c):
            return ((b[0] - a[0])*(c[1] - a[1]) - (b[1] - a[1])*(c[0] - a[0])) > 0
        
        def stretcher_slnt(h, xy, angle, i, p):
            if abs(angle) in [90, 270]:
                return p
            x0, y0 = xy
            ra = math.radians(90+angle)
            ydsc = y0 + (0 - x0) / math.tan(ra)
            yasc = y0 + (p.ambit().point("E").x - x0) / math.tan(ra)
            p0 = (0, ydsc)
            p1 = (p.ambit().point("E").x, yasc)

            np = (p.flatten(flatten) if flatten else p).nonlinear_transform(lambda x,y: (x, y if not is_left(p0, p1, (x, y)) else y + h))
            if debug:
                (np
                    .record(_PenClass()
                        .line([p0, p1])
                        .outline())
                    .record(_PenClass()
                        .moveTo((x0+50/2, y0+50/2))
                        .dots(radius=50)))
            return np
        
        for k, v in kwargs.items():
            if len(v) == 3:
                d[k] = (0, partial(stretcher_slnt, v[0], v[1], v[2]))
            elif len(v) == 2:
                d[k] = (0, partial(stretcher, v[0], v[1]))
            elif len(v) == 1:
                pass
        return d


def offset(x, y, ox, oy):
    return (x + ox, y + oy)


class StyledString(FittableMixin):
    """
    Lowest-level vectorized typesetting class
    """
    def __init__(self, text:str, style:Style):
        self.text_info = TextInfo(text)
        self.text = text
        self.setStyle(style)
        if self.style.lang:
            self.text_info.languageOverride = self.style.lang
        self.resetGlyphRun()
    
    def setStyle(self, style):
        self.style = style
        # these will change based on fitting, so we make copies
        self.fontSize = self.style.fontSize
        self.tracking = self.style.tracking
        self.features = self.style.features.copy()
        self.variations = self.style.variations.copy()
    
    def resetGlyphRun(self):
        self.glyphs = self.style.font.font.getGlyphRunFromTextInfo(self.text_info, addDrawings=False, features=self.features, varLocation=self.variations)
        #self.glyphs = self.style.font.font.getGlyphRun(self.text, features=self.features, varLocation=self.variations)
        x = 0
        for glyph in self.glyphs:
            glyph.frame = Rect(x+glyph.dx, glyph.dy, glyph.ax, self.style.capHeight)
            x += glyph.ax
        self.getGlyphFrames()
    
    def trackFrames(self, space_width=0):
        t = self.tracking
        x_off = 0
        for idx, g in enumerate(self.glyphs):
            g.frame = g.frame.offset(x_off, 0)
            x_off += t
            if self.style.mods and g.name in self.style.mods:
                x_off += self.style.mods[g.name][0]
            if self.style.space and g.name.lower() == "space":
                x_off += (self.style.space - space_width)
    
    def adjustFramesForPath(self):
        for idx, g in enumerate(self.glyphs):    
            if self.style.xShift:
                try:
                    g.frame.x += self.style.xShift[idx]
                except:
                    g.frame.x += self.style.xShift
                    pass
    
    def getGlyphFrames(self):
        if self.style.kern_pairs:
            last_gn = None
            for idx, glyph in enumerate(self.glyphs):
                gn = glyph.name

                for chars, kp in self.style.kern_pairs.items():
                    try:
                        l = kp[0]
                        adv = kp[1]
                    except TypeError:
                        l = kp
                        adv = 0
                    if isinstance(chars, str):
                        a, b = chars.split("/")
                    else:
                        a, b = chars
                    if gn == b and last_gn == a:
                        kern_shift = l
                        if kern_shift != 0:
                            for glyph in self.glyphs[idx:]:
                                glyph.frame.x += kern_shift
                last_gn = gn
        
        space_width = 0
        for glyph in self.glyphs:
            if glyph.name == "space" and self.style.space and self.style.space > 0:
                space_width = glyph.frame.w
                glyph.frame.w = self.style.space

        if self.style.trackingMode == 1:
            self.trackFrames(space_width=space_width)

        for glyph in self.glyphs:
            #print(glyph, glyph.frame)
            glyph.frame = glyph.frame.scale(self.scale())

        if self.style.trackingMode == 0:
            self.trackFrames()

        self.adjustFramesForPath()
    
    def scale(self):
        return self.fontSize / self.style.font.font.shaper.face.upem
    
    def width(self): # size?
        w = self.glyphs[-1].frame.point("SE").x # TODO need to scale?
        #return w * self.scale()
        return w
        return self.getGlyphFrames()[-1].frame.point("SE").x
    
    def height(self):
        return self.style.capHeight*self.scale()
    
    def textContent(self):
        return self.text
    
    def fitField(self, field, value):
        if field == "tracking":
            self.tracking = value
        elif field == "wdth":
            self.variations["wdth"] = value
        elif field == "fontSize":
            self.fontSize = value
    
    def binaryFit(self, width, field, minv, maxv, tries):
        midv = (maxv-minv)*0.5+minv
        self.fitField(field, maxv)
        self.resetGlyphRun()
        maxw = self.width()
        self.fitField(field, midv)
        self.resetGlyphRun()
        midw = self.width()
        self.fitField(field, minv)
        self.resetGlyphRun()
        minw = self.width()
        if abs(maxw - midw) < 0.5:
            #print(self.text, ">>>", tries)
            return
        if width > midw:
            self.binaryFit(width, field, midv, maxv, tries+1)
        else:
            self.binaryFit(width, field, minv, midv, tries+1)
        #return super().fit(width)
    
    def testWidth(self, width, field, minv, maxv):
        self.resetGlyphRun()
        w = self.width()
        if w == width:
            print("VERY RARE")
            return True
        elif w < width: # too small, which means we know it'll fit based on this property
            self.binaryFit(width, field, minv, maxv, 0)
            return True
        else: # too big, so we maintain current value & let the caller know
            return False
    
    def _fit(self, width):
        if isinstance(width, Rect):
            width = width.w

        continuing = True
        failed = False

        if self.style.tracking > 0:
            self.tracking = 0
            if self.testWidth(width, "tracking", 0, self.style.tracking):
                continuing = False
        if continuing:
            minwdth = self.style.variationLimits.get("wdth", 1)
            currentwdth = self.style.variations.get("wdth", 1)
            self.variations["wdth"] = minwdth
            if not self.testWidth(width, "wdth", minwdth, currentwdth):
                #self.tracking = self.style.trackingLimit
                if not self.testWidth(width, "tracking", self.style.trackingLimit, min(self.style.tracking, 0)):
                    if self.style.varyFontSize:
                        self.fontSize = 10
                        if not self.testWidth(width, "fontSize", 10, self.style.fontSize):
                            self.variations["wdth"] = minwdth
                            failed = True
                    else:
                        self.variations["wdth"] = minwdth
                        failed = True
        if failed:
            print("FAILED TO FIT >>>", self.text, self.width(), width)
        return self

    def shrink(self):
        #print(self.text, self.variations)
        adjusted = False
        default_step = 1
        if self.tracking > 0 and self.tracking > self.style.trackingLimit:
            self.tracking -= self.style.increments.get("tracking", default_step)
            adjusted = True
        else:
            for k, v in self.style.variationLimits.items():
                if self.variations[k] > self.style.variationLimits[k]:
                    self.variations[k] -= self.style.increments.get(k, default_step)
                    adjusted = True
                    break
        if not adjusted and self.tracking > self.style.trackingLimit:
            self.tracking -= self.style.increments.get("tracking", default_step)
            adjusted = True
        if not adjusted and self.style.varyFontSize:
            self.fontSize -= 1
            self.tracking = self.style.tracking
            adjusted = True
        if not adjusted and self.style.narrower:
            self.setStyle(self.style.narrower)
            adjusted = True
        if not adjusted and self.style.preventHwid == False and "hwid" not in self.features:
            self.features["hwid"] = True
            self.tracking = self.style.tracking # reset to widest
            self.resetGlyphRun()
            #self.glyphs = self.hb.glyphs(self.variations, self.features)
            adjusted = True
        self.resetGlyphRun()
        return adjusted
    
    def scalePenToStyle(self, glyph, in_pen):
        s = self.scale()
        t = Transform()
        try:
            bs = self.style.baselineShift[idx]
        except:
            bs = self.style.baselineShift
        
        if callable(bs):
            t = t.translate(0, bs(idx))
        else:
            try:
                t = t.translate(0, bs[idx])
            except:
                try:
                    t = t.translate(0, bs)
                except:
                    pass
        t = t.scale(s)
        t = t.translate(glyph.frame.x/self.scale(), glyph.frame.y/self.scale())
        #t = t.translate(0, bs)
        out_pen = _PenClass()
        tp = TransformPen(out_pen, (t[0], t[1], t[2], t[3], t[4], t[5]))
        ip = _PenClass().record(in_pen)
        if self.style.mods and glyph.name in self.style.mods:
            w, mod = self.style.mods[glyph.name]
            mod(-1, ip)
        ip.replay(tp)
        if self.style.rotate:
            out_pen.rotate(self.style.rotate)
        
        # TODO this shouldn't be necessary
        if True:
            valid_values = []
            for (move, pts) in out_pen.value:
                if move != "addComponent":
                    valid_values.append((move, pts))
            out_pen.value = valid_values

        return out_pen
    
    def _emptyPenWithAttrs(self):
        #attrs = dict(fill=self.style.fill)
        #if self.style.stroke:
        #    attrs["stroke"] = dict(color=self.style.stroke, weight=self.style.strokeWidth)
        dp = _PenClass().f(self.style.fill)
        if self.style.stroke:
            dp.s(self.style.stroke).sw(self.style.strokeWidth)
        return dp

    def pens(self, frame=True) -> DraftingPens:
        """
        Vectorize text into a ``DATPens``, such that each glyph (or ligature) is represented by a single `DATPen` (or a ``DATPens`` in the case of a color font, which will then nest a `DATPen` for each layer of that color glyph)
        """
        self.resetGlyphRun()
        self.style.font.font.addGlyphDrawings(self.glyphs, colorLayers=True)
        
        pens = _PensClass()
        for idx, g in enumerate(self.glyphs):
            dp_atom = self._emptyPenWithAttrs()
            if len(g.glyphDrawing.layers) == 1:
                dp_atom.value = self.scalePenToStyle(g, g.glyphDrawing.layers[0][0]).value
                dp_atom.typographic = True
                dp_atom.addFrame(g.frame)
                dp_atom.glyphName = g.name
                if self.style.removeOverlap:
                    dp_atom.removeOverlap()
            else:
                dp_atom = _PensClass()
                dp_atom.layered = True
                for layer in g.glyphDrawing.layers:
                    dp_layer = self._emptyPenWithAttrs()
                    #dp_layer.value = layer[0].value
                    dp_layer.value = self.scalePenToStyle(g, layer[0]).value
                    dp_layer.f(self.style.font.font.colorPalettes[self.style.palette][layer[1]])
                    dp_atom += dp_layer
                dp_atom.addFrame(g.frame, typographic=True)
                dp_atom.glyphName = g.name
            pens.append(dp_atom, allow_blank=self.style.include_blanks)

        if self.style.reverse:
            pens.reversePens()
        
        for k, v in self.style.data.items():
            pens.data[k] = v

        return pens

    def pen(self, frame=True) -> DraftingPen:
        """
        Vectorize all text into single ``DATPen``
        """
        return self.pens(frame=frame).pen()

class SegmentedString(FittableMixin):
    def __init__(self, text, styles):
        self.text_info = TextInfo(text)
        self.strings = []
        self.segment_data = []
        for segmentText, segmentScript, segmentBiDiLevel, firstCluster in self.text_info._segments:
            self.strings.append(StyledString(segmentText, styles[segmentScript]))
            cluster_data = []
            for index, char in enumerate(segmentText, firstCluster):
                cluster_data.append(
                    dict(index=index, char=char, unicode=f"U+{ord(char):04X}",
                        unicodeName=unicodedata.name(char, "?"), script=segmentScript,
                        bidiLevel=segmentBiDiLevel, dir=["LTR", "RTL"][segmentBiDiLevel % 2])
                )
            self.segment_data.append(cluster_data)
    
    def width(self):
        return sum([s.width() for s in self.strings])
    
    def height(self):
        return max([s.height() for s in self.strings])
    
    def textContent(self):
        return "-".join([s.textContent() for s in self.strings])

    def shrink(self):
        adjusted = False
        for s in self.strings:
            adjusted = s.shrink() or adjusted
        return adjusted

    def pens(self):
        pens = _PensClass()
        x_off = 0
        for s in self.strings:
            dps = s.pens(frame=True)
            if dps.layered:
                pens.layered = True
            dps.translate(x_off, 0)
            pens.extend(dps._pens)
            x_off += dps.ambit().w
        return pens
        #return DATPens([s.pens(frame=True) for s in self.strings])


from coldtype.renderable.renderable import Keylayer, Overlay, Action, RenderPass, renderable, skia_direct, drawbot_script, svgicon, glyph, fontpreview, iconset, SkiaPen
from coldtype.renderable.animation import animation, FFMPEGExport, drawbot_animation


from enum import Enum
from pathlib import Path
import json, re, base64
from coldtype import hsl, Action, Keylayer, Point, Rect, DATPen, Overlay
from typing import Callable, List
from time import sleep

try:
    import skia
except ImportError:
    skia = None

try:
    import glfw
except ImportError:
    glfw = None

class RendererStateEncoder(json.JSONEncoder):
    def default(self, o):
        return {
            "controller_values": o.controller_values
        }


class Mods():
    def __init__(self):
        self.reset()
    
    def update(self, mods):
        if mods & glfw.MOD_SUPER:
            self.super = True
        if mods & glfw.MOD_SHIFT:
            self.shift = True
        if mods & glfw.MOD_ALT:
            self.alt = True
        if mods & glfw.MOD_CONTROL:
            self.ctrl = True
    
    def reset(self):
        self.super = False
        self.shift = False
        self.ctrl = False
        self.alt = False
    
    def __repr__(self):
        return f"su:{self.super};sh:{self.shift};alt:{self.alt};ctrl:{self.ctrl}"


class InputHistoryItem():
    def __init__(self, position, action, keylayer, midi):
        self.position = position
        self.action = action
        self.keylayer = keylayer
        self.midi = midi
        self.idx = -1
    
    def __repr__(self):
        return f"InputHistoryItem({self.idx}/{self.action}/{self.position})"


class InputHistoryGroup():
    def __init__(self, action, keylayer, items):
        self.action = action
        self.keylayer = keylayer
        self.items = []
    
    def append(self, item):
        self.items.append(item)
    
    def __len__(self):
        return len(self.items)
    
    def __getitem__(self, index):
        return self.items[index]
    
    def position(self):
        if len(self.items) > 0:
            return self.items[0].position
    
    def start(self):
        if len(self.items) > 0:
            return self.items[0].idx
    
    def end(self):
        if len(self.items) > 0:
            return self.items[-1].idx
    
    def __repr__(self):
        return f"InputHistoryGroup:{self.action}:{self.start()}-{self.end()}"


class InputHistory():
    def __init__(self):
        self.history = []

    def clear(self) -> bool:
        """Return True if something was cleared"""
        if len(self.history) > 0:
            self.history = []
            return True
        else:
            self.history = []
            return False
    
    def empty(self):
        return len(self.history) == 0
    
    def record(self, item:InputHistoryItem):
        item.idx = len(self.history)
        self.history.append(item)
    
    def last(self, action=None) -> InputHistoryItem:
        for i in reversed(self.history):
            if not action:
                return i
            elif i.action == action:
                return i
            elif callable(action):
                if action(i.action):
                    return i
    
    def undo(self):
        strokes = self.strokes()
        if strokes:
            last_stroke = strokes[-1]
            self.history = self.history[:last_stroke.start()]
    
    def strokes(self, filterfn:Callable[[InputHistoryGroup], bool]=lambda g: True) -> List[InputHistoryGroup]:
        xs = []
        for i in self.history:
            if i.action == "down":
                xs.append(InputHistoryGroup("down", i.keylayer, []))
            elif i.action == "up":
                xs.append(InputHistoryGroup("up", i.keylayer, []))
            elif i.action == "cmd":
                xs.append(InputHistoryGroup("cmd", i.keylayer, []))
            
            if len(xs) == 0 and i.action == "move":
                continue
            else:
                xs[-1].append(i)
        return [x for x in xs if filterfn(x)]
    
    def downstrokes(self):
        return self.strokes(lambda g: g.action == "down")
    
    def upstrokes(self):
        return self.strokes(lambda g: g.action == "up")
    
    def moved(self) -> bool:
        last_down = self.last("down")
        last_up = self.last("up")
        if last_down and last_up and last_down.idx < last_up.idx:
            return last_down.position != last_up.position
        return False
    
    def __getitem__(self, index):
        if len(self.history) > 0:
            return self.history[index]


class RendererState():
    def __init__(self, renderer):
        self.renderer = renderer
        self.previewing = False
        self.preview_scale = 1
        self.controller_values = {}
        self.keylayer_shifting = False
        self.keylayer = Keylayer.Default
        self.keybuffer = []
        self.overlays = {}
        #self.needs_display = 0
        self.request = None
        self.callback = None
        self.cmd = None
        self.text = ""
        self.clear_text = False
        self.arrow = None
        self.mods = Mods()
        self.record_mouse_moves = True
        self.input_history:InputHistory = InputHistory()
        self.xray = True
        self.selection = [0]
        self.zoom = 1
        self._frame_offsets = {}
        self._initial_frame_offsets = {}
        self.canvas = None
        self._last_filepath = None
        self.watch_soft_mods = {}
        self.watch_mods = {}
        self.external_url = None
        self.cv2caps = {}
        self.capturing_previews = False
        self.capturing_previews_once = False
        self.captured_previews = []
        self.reset()
    
    def reset(self, ignore_current_state=False):
        if self.filepath == self._last_filepath and not ignore_current_state:
            return
        
        if self.filepath:
            self._last_filepath = self.filepath
            try:
                deserial = json.loads(self.filepath.read_text())
                cv = deserial.get("controller_values")
                if cv:
                    self.controller_values = cv
            except json.decoder.JSONDecodeError:
                self.controller_values = {}
            except FileNotFoundError:
                self.controller_values = {}
    
    def clear(self):
        if self.filepath:
            self.filepath.write_text("")
        self.reset()
    
    @property
    def filepath(self):
        if self.renderer.filepath:
            return Path(str(self.renderer.filepath).replace(".py", "") + "_state.json")
        else:
            return None
    
    @property
    def midi(self):
        return self.controller_values
    
    def persist(self):
        if self.filepath:
            print("Saving Controller State...")
            self.filepath.write_text(RendererStateEncoder().encode(self))
        else:
            print("No source; cannot persist state")
    
    def record_mouse(self, pos, action):
        new_mouse = pos.scale(1/self.preview_scale)
        self.input_history.record(
            InputHistoryItem(new_mouse, action, self.keylayer, self.controller_values.copy()))
        if action == "down":
            return Action.PreviewStoryboard
        elif action == "up" and self.input_history.moved():
            return Action.PreviewStoryboard
        elif action == "move" and self.keylayer == Keylayer.Editing:
            return Action.PreviewStoryboard
    
    def on_mouse_button(self, pos, btn, action, mods):
        if btn == 1 and action == glfw.PRESS:
            self.input_history.clear()
            return Action.PreviewStoryboard
        
        self.mods.update(mods)
        if action == glfw.PRESS:
            return self.record_mouse(pos, "down")
        elif action == glfw.RELEASE:
            return self.record_mouse(pos, "up")
    
    def on_mouse_move(self, pos):
        if self.record_mouse_moves:
            return self.record_mouse(pos, "move")
    
    @property
    def mouse_history(self):
        strokes = self.input_history.downstrokes()
        strokes = [[p.position for p in s.items] for s in strokes]
        return strokes
    
    @property
    def mouse_down(self):
        strokes = self.input_history.strokes()
        if len(strokes) > 0 and strokes[-1][0] == "down":
            return True
        else:
            return False
    
    @property
    def mouse(self):
        item = self.input_history.last(lambda a: a in ["up", "down", "move"])
        if item:
            return item.position
    
    def mod_preview_scale(self, inc, absolute=0):
        if absolute > 0:
            ps = absolute
        else:
            ps = self.preview_scale + inc
        self.preview_scale = max(0.1, min(5, ps))
    
    def shape_selection(self):
        # TODO could be an arbitrary lasso-style thing?
        if self.mouse_history:
            start = self.mouse_history[-1][0]
            end = self.mouse_history[-1][-1]
            rect = Rect.FromPoints(start, end)
            return DATPen().rect(rect)
        
    def polygon_selection(self, clear_on="save"):
        mh = self.mouse_history
        if not mh:
            return False, DATPen()
        
        polygon = DATPen()
        polygon.line([m[-1] for m in mh])
        polygon.f(None).s(hsl(0.75, s=1, a=0.5)).sw(5)
        #polygon.closePath()

        if self.cmd == clear_on:
            self.input_history.clear()
            return True, polygon
        else:
            return False, polygon
    
    def on_character(self, codepoint):
        cc = chr(codepoint)
        if self.keylayer == Keylayer.Cmd or self.keylayer == Keylayer.Text:
            self.keybuffer.append(cc)
            return Action.PreviewStoryboard
        
        elif self.keylayer == Keylayer.Editing:
            if cc == "f":
                return
            if re.match(r"[1-9]{1}", cc):
                self.zoom = int(cc)
            elif cc == "=":
                self.zoom += 0.25
            elif cc == "-":
                self.zoom -= 0.25
            #else:
            #    self.cmd = cc
            self.zoom = max(0.25, min(10, self.zoom))
            return Action.PreviewStoryboard
        #self.needs_display = 1
    
    def on_key(self, win, key, scan, action, mods):
        if action != glfw.PRESS and action != glfw.REPEAT:
            return
        
        self.mods.update(mods)
        
        if key == glfw.KEY_ENTER:
            cmd = "".join(self.keybuffer)
            if self.keylayer == Keylayer.Cmd:
                self.cmd = cmd
            elif self.keylayer == Keylayer.Text:
                self.text = cmd
            self.keybuffer = []
            #print(">>> KB-CMD:", cmd)
            if mods & glfw.MOD_SUPER:
                # keep editing
                pass
            else:
                self.exit_keylayer()
            return Action.PreviewStoryboard
        elif key == glfw.KEY_BACKSPACE:
            if len(self.keybuffer) > 0:
                self.keybuffer = self.keybuffer[:-1]
                return Action.PreviewStoryboard
        elif key == glfw.KEY_ESCAPE:
            if self.keylayer == Keylayer.Editing:
                cleared = self.input_history.clear()
                if cleared:
                    return Action.PreviewStoryboard
            
            self.exit_keylayer()
            return Action.PreviewStoryboard

        elif self.keylayer == Keylayer.Editing:
            if key == glfw.KEY_D:
                self.exit_keylayer()
                return Action.PreviewStoryboard
            elif key in [glfw.KEY_UP, glfw.KEY_DOWN, glfw.KEY_LEFT, glfw.KEY_RIGHT]:
                if key == glfw.KEY_UP:
                    self.arrow = [0, 1]
                elif key == glfw.KEY_DOWN:
                    self.arrow = [0, -1]
                elif key == glfw.KEY_LEFT:
                    self.arrow = [-1, 0]
                elif key == glfw.KEY_RIGHT:
                    self.arrow = [1, 0]
                if mods & glfw.MOD_SHIFT:
                    for idx, a in enumerate(self.arrow):
                        self.arrow[idx] = a * 10
                return Action.PreviewStoryboard
            elif key == glfw.KEY_E and self.keylayer == Keylayer.Editing:
                self.exit_keylayer()
                return Action.PreviewStoryboard
            elif key == glfw.KEY_F and self.keylayer == Keylayer.Editing:
                self.xray = not self.xray
                return Action.PreviewStoryboard
            elif key == glfw.KEY_Z and self.keylayer == Keylayer.Editing:
                self.input_history.undo()
                return Action.PreviewStoryboard
            elif key == glfw.KEY_B and self.keylayer == Keylayer.Editing:
                self.input_history.record(InputHistoryItem(key, "cmd", self.keylayer, self.controller_values.copy()))
            elif key == glfw.KEY_C and self.keylayer == Keylayer.Editing:
                self.input_history.record(InputHistoryItem(key, "cmd", self.keylayer, self.controller_values.copy()))

        
        if key == glfw.KEY_S and mods & glfw.MOD_SUPER:
            self.cmd = "save"
            return Action.PreviewStoryboard
        return
    
    def exit_keylayer(self):
        self.keylayer = Keylayer.Default
        self.keybuffer = []
    
    def draw_keylayer(self, canvas, rect, typeface):
        canvas.save()

        if self.keylayer == Keylayer.Cmd:
            canvas.drawRect(
                skia.Rect.MakeXYWH(0, rect.h-50, rect.w, 50),
                skia.Paint(
                    AntiAlias=True,
                    Color=hsl(0.9, l=0.25, a=0.85).skia()))
            canvas.drawString(
                "".join(self.keybuffer),
                10,
                rect.h-14,
                skia.Font(typeface, 30),
                skia.Paint(
                    AntiAlias=True,
                    Color=skia.ColorWHITE))
        
        elif self.keylayer == Keylayer.Text:
            canvas.drawRect(
                skia.Rect.MakeXYWH(0, 0, rect.w, 50),
                skia.Paint(
                    AntiAlias=True,
                    Color=hsl(0.7, l=0.25, a=0.85).skia()))
            canvas.drawString(
                "".join(self.keybuffer),
                10,
                34,
                skia.Font(typeface, 30),
                skia.Paint(
                    AntiAlias=True,
                    Color=skia.ColorWHITE))
        
        elif self.keylayer == Keylayer.Editing:
            canvas.drawRect(skia.Rect(0, 0, 50, 50), skia.Paint(AntiAlias=True, Color=hsl(0.95, l=0.5, a=0.75).skia()))
        canvas.restore()
    
    def increment_selection(self, amount, limit):
        if len(self.selection) == 1:
            if amount == 0:
                return
            si = self.selection[0]
            si += amount
            if si >= limit:
                si = 0
            elif si < 0:
                si = limit - 1
            self.selection[0] = si
        else:
            print("invalid")
    
    def add_frame_offset(self, key, offset):
        if key in self._frame_offsets:
            offsets = self._frame_offsets[key]
            initials = self._initial_frame_offsets[key]
            offsets.append(offset)
            initials.append(offset)
        else:
            self._frame_offsets[key] = [offset]
            self._initial_frame_offsets[key] = [offset]
    
    def get_frame_offsets(self, key):
        return self._frame_offsets.get(key, [0])
    
    def adjust_all_frame_offsets(self, adj, absolute=False):
        if absolute:
            for k, v in self._frame_offsets.items():
                for i, o in enumerate(v):
                    v[i] = adj
        else:
            for k, v in self._frame_offsets.items():
                for i, o in enumerate(v):
                    v[i] = o + adj
    
    def adjust_keyed_frame_offsets(self, key, fn):
        offs = self.get_frame_offsets(key)
        for i, o in enumerate(offs):
            offs[i] = fn(i, o)
    
    def read_text(self, clear=False):
        self.clear_text = clear
        if self.keybuffer:
            return "".join(self.keybuffer)
        elif self.text:
            return self.text
        else:
            return None
    
    def reset_keystate(self):
        self.cmd = None
        if self.clear_text:
            self.text = ""
            self.clear_text = False
        self.arrow = None
        self.mods.reset()
        self.watch_mods = {}
        self.watch_soft_mods = {}
    
    def toggle_overlay(self, overlay):
        v = not self.overlays.get(overlay, False)
        if not v:
            if overlay in self.overlays:
                del self.overlays[overlay]
        else:
            self.overlays[overlay] = True
    
    def notify_external(self, idx):
        from websocket import create_connection
        ws = None
        try:
            ws = create_connection(self.external_url)
            print("> open", idx)
            ws.send(json.dumps({"rendered": idx}))
        except ConnectionRefusedError:
            print("!!! Could not connect to websocket", self.external_url)
        if ws:
            print("< close", idx)
            ws.close()
        return self


import math
from fontTools.pens.recordingPen import RecordingPen, replayRecording
from fontTools.misc.bezierTools import calcCubicArcLength, splitCubicAtT, calcQuadraticArcLength
from coldtype.geometry import Rect, Point, Line


def raise_quadratic(start, a, b):
    c0 = start
    c1 = (c0[0] + (2/3)*(a[0] - c0[0]), c0[1] + (2/3)*(a[1] - c0[1]))
    c2 = (b[0] + (2/3)*(a[0] - b[0]), b[1] + (2/3)*(a[1] - b[1]))
    c3 = (b[0], b[1])
    return [c1, c2, c3]

__length_cache = {}
__split_cache = {}

def splitCubicAtT_cached(a, b, c, d, t):
    global __split_cache
    abcdt = (a, b, c, d, t)
    sc = __split_cache.get(abcdt)
    if sc:
        return sc
    else:
        s = splitCubicAtT(a, b, c, d, t)
        __split_cache[abcdt] = s
        return s

def calcCubicArcLength_cached(a, b, c, d):
    #return calcCubicArcLength(a, b, c, d)
    global __length_cache
    abcd = (a, b, c, d)
    lc = __length_cache.get(abcd)
    if lc:
        return lc
    else:
        l = calcCubicArcLength(a, b, c, d)
        __length_cache[abcd] = l
        return l

class CurveCutter():
    def __init__(self, g, inc=0.0015):
        #self.pen = g#RecordingPen()
        self.pen = g.copy()
        #g.draw(self.pen)
        #self.pen.ensure_fully_closed_path()
        self.inc = inc
        self.length = self.calcCurveLength()
    
    def calcCurveLength(self):
        length = 0
        for i, (t, pts) in enumerate(self.pen.value):
            if t == "curveTo":
                p1, p2, p3 = pts
                p0 = self.pen.value[i-1][-1][-1]
                length += calcCubicArcLength_cached(p0, p1, p2, p3)
            elif t == "qCurveTo":
                p1, p2 = pts
                p0 = self.pen.value[i-1][-1][-1]
                length += calcQuadraticArcLength(p0, p1, p2)
            elif t == "lineTo":
                p0 = self.pen.value[i-1][-1][-1]
                p1, = pts
                l = Line(p0, p1)
                length += l.length()
        return length

    def subsegment(self, start=None, end=None):
        global __cut_cache
        inc = self.inc
        length = self.length
        ended = False
        _length = 0
        out = []
        for i, (t, pts) in enumerate(self.pen.value):
            if t == "curveTo":
                p1, p2, p3 = pts
                p0 = self.pen.value[i-1][-1][-1]
                length_arc = calcCubicArcLength_cached(p0, p1, p2, p3)
                if _length + length_arc < end:
                    _length += length_arc
                else:
                    t = inc
                    tries = 0
                    while not ended:
                        a, b = splitCubicAtT_cached(p0, p1, p2, p3, t)
                        length_a = calcCubicArcLength_cached(*a)
                        if _length + length_a > end:
                            ended = True
                            out.append(("curveTo", a[1:]))
                        else:
                            t += inc
                            tries += 1
            elif t == "lineTo":
                p0 = self.pen.value[i-1][-1][-1]
                p1, = pts
                l = Line(p0, p1)
                length_line = l.length()
                if ended:
                    pass
                elif _length + length_line < end:
                    _length += length_line
                else:
                    res = l.tpx(end - _length)
                    out.append(("lineTo", (res,)))
                    ended = True
            
            if not ended:
                out.append((t, pts))
    
        if out[-1][0] != "endPath":
            out.append(("endPath",[]))
        return out

    def subsegmentPoint(self, start=0, end=1):
        inc = self.inc
        subsegment = self.subsegment(start=start, end=end)
        try:
            t, (a, b, c) = subsegment[-2]
            tangent = math.degrees(math.atan2(c[1] - b[1], c[0] - b[0]) + math.pi*.5)
            return Point(c), tangent
        except ValueError:
            return None, None


try:
    import drawBot as db
except:
    pass

from coldtype.pens.draftingpen import DraftingPen
from coldtype.pens.draftingpens import DraftingPens
from coldtype.geometry import Rect, Edge, Point
from coldtype.pens.drawablepen import DrawablePenMixin
from coldtype.color import Color, Gradient


def get_image_rect(src):
    w, h = db.imageSize(str(src))
    return Rect(0, 0, w, h)


class DrawBotPen(DrawablePenMixin, DraftingPen):
    def __init__(self, dat, rect=None):
        super().__init__()
        self.rect = rect
        self.dat = dat
        self.bp = db.BezierPath()
        self.dat.replay(self.bp)
    
    def fill(self, color):
        if color:
            if isinstance(color, Gradient):
                self.gradient(color)
            elif isinstance(color, Color):
                db.fill(color.r, color.g, color.b, color.a)
        else:
            db.fill(None)
    
    def stroke(self, weight=1, color=None, dash=None):
        db.strokeWidth(weight)
        if dash:
            db.lineDash(dash)
        if color:
            if isinstance(color, Gradient):
                pass # possible?
            elif isinstance(color, Color):
                db.stroke(color.r, color.g, color.b, color.a)
        else:
            db.stroke(None)
        
    def image(self, src=None, opacity=1, rect=None, rotate=0, repeating=False, scale=True):
        bounds = self.dat.bounds()
        src = str(src)
        if not rect:
            rect = bounds
        try:
            img_w, img_h = db.imageSize(src)
        except ValueError:
            print("DrawBotPen: No image")
            return
        x = bounds.x
        y = bounds.y
        if repeating:
            x_count = bounds.w / rect.w
            y_count = bounds.h / rect.h
        else:
            x_count = 1
            y_count = 1
        _x = 0
        _y = 0
        while x <= (bounds.w+bounds.x) and _x < x_count:
            _x += 1
            while y <= (bounds.h+bounds.y) and _y < y_count:
                _y += 1
                with db.savedState():
                    r = Rect(x, y, rect.w, rect.h)
                    #db.fill(1, 0, 0.5, 0.05)
                    #db.oval(*r)
                    if scale == True:
                        db.scale(rect.w/img_w, center=r.point("SW"))
                    elif scale:
                        try:
                            db.scale(scale[0], scale[1], center=r.point("SW"))
                        except TypeError:
                            db.scale(scale, center=r.point("SW"))
                    db.rotate(rotate)
                    db.image(src, (r.x, r.y), alpha=opacity)
                y += rect.h
            y = 0
            x += rect.w
    
    def shadow(self, clip=None, radius=10, alpha=0.3, color=Color.from_rgb(0,0,0,1)):
        if clip:
            cp = DraftingPen(clip).f(None)
            bp = db.BezierPath()
            cp.replay(bp)
            db.clipPath(bp)
        #elif self.rect:
        #    cp = DATPen(fill=None).rect(self.rect).xor(self.dat)
        #    bp = db.BezierPath()
        #    cp.replay(bp)
        #    db.clipPath(bp)
        db.shadow((0, 0), radius*3, list(color.with_alpha(alpha)))

    def gradient(self, gradient):
        stops = gradient.stops
        db.linearGradient(stops[0][1], stops[1][1], [list(s[0]) for s in stops], [0, 1])
    
    def draw(self, scale=1, style=None):
        if hasattr(self.dat, "_pens"):
            for p in self.dat._pens:
                DrawBotPen(p, rect=self.rect).draw(scale=scale)
        else:
            with db.savedState():
                db.scale(scale)
                for attrs, attr in self.findStyledAttrs(style):
                    self.applyDATAttribute(attrs, attr)
                db.drawPath(self.bp)
        return self
    
    def draw_with_filters(self, rect, filters):
        im = db.ImageObject()
        with im:
            db.size(*rect.wh())
            self.draw()
        for filter_name, filter_kwargs in filters:
            getattr(im, filter_name)(**filter_kwargs)
        x, y = im.offset()
        db.image(im, (x, y))
        return self


import re, math
from coldtype.geometry import Rect, Edge, Point


def parse_line(d, line):
    parts = re.split(r"\s", line)
    reified = []
    for p in parts:
        if p == "auto" or p == "a":
            reified.append("auto")
        elif "%" in p:
            reified.append(float(p.replace("%", ""))/100 * d)
        else:
            reified.append(float(p))
    remaining = d - sum([0 if r == "auto" else r for r in reified])
    if not float(remaining).is_integer():
        remaining = round(remaining)
        #raise Exception("floating parse")
    auto_count = reified.count("auto")
    if auto_count > 0:
        auto_d = remaining / auto_count
        auto_ds = [auto_d] * auto_count
        if not auto_d.is_integer():
            auto_d_floor = math.floor(auto_d)
            leftover = remaining - auto_d_floor * auto_count
            for aidx, ad in enumerate(auto_ds):
                if leftover > 0:
                    auto_ds[aidx] = auto_d_floor + 1
                    leftover -= 1
                else:
                    auto_ds[aidx] = auto_d_floor
            #print(auto_ds)
            #print("NO", auto_d - int(auto_d))
    res = []
    for r in reified:
        if r == "auto":
            res.append(auto_ds.pop())
        else:
            res.append(r)
    return res
    #return [auto_d if r == "auto" else r for r in reified]


def union_rect(r1, r2):
    ox = min(r1.x, r2.x)
    oy = min(r1.y, r2.y)
    ex = max(r1.x+r1.w, r2.x+r2.w)
    ey = max(r1.y+r1.h, r2.y+r2.h)
    return Rect(ox, oy, ex-ox, ey-oy)
    
class Grid():
    def __init__(self, r, columns="auto", rows="auto", areas=None):
        self._rect = r
        self.columns = columns
        self.rows = rows
        self.areas = areas
        self.cells = None
        self.keyed = None
        self.update()
    
    def __repr__(self):
        return f"Grid({list(self.keyed.keys())})"
    
    def clone(self, other_grid):
        self.columns = other_grid.columns
        self.rows = other_grid.rows
        self.areas = other_grid.areas
        return self
    
    @property
    def rect(self):
        return self._rect
    
    @rect.setter
    def rect(self, rect):
        self._rect = rect
        self.update()
        return self

    def key(self):
        if isinstance(self._rect, str):
            return self._rect
    
    def update(self):
        if self.key():
            return

        # TODO should be read from a configurable dict
        # (on a subclass of grid?)
        self.columns = self.columns.replace("$CLH", "36")
        self.rows = self.rows.replace("$CLH", "36")

        cg, bs = self.calc_grid(self._rect, self.columns, self.rows, self.areas)
        self.borders = bs
        if self.areas:
            self.keyed = cg
            self.cells = None
        else:
            self.cells = cg
            self.keyed = None
        
    def __getitem__(self, key):
        if self.cells:
            if isinstance(key, int):
                return self.cells[key]
            else:
                raise Exception("Must query cell-grid with indices")
        elif self.keyed:
            if isinstance(key, str):
                return self.keyed[key]
            else:
                print(">>>>>>>>>>", self, key)
                raise Exception("Must query area-grid with strings")

    def calc_grid(self, r, columns, rows, areas):
        cs = parse_line(r.w, columns)
        rs = parse_line(r.h, rows)
        if areas:
            areas = [a.strip().split(" ") for a in areas.split("/")]
        _grid = []
        keyed = {}
        borders = []
        for idx, rr in enumerate(r.subdivide(rs[:-1], "mxy")):
            cells = rr.subdivide(cs[:-1], "mnx")
            _grid.extend(cells)
            if areas:
                _keyed = {}
                last_area = None
                jdx = 0
                if idx >= len(areas):
                    continue
                for ra in areas[idx]:
                    if not ra:
                        continue
                    if ra == "||":
                        borders.append([last_area, Edge.MaxX, 1])
                        continue
                    elif ra == "|":
                        borders.append([last_area, Edge.MaxX, 0])
                        continue
                    elif ra == "__":
                        borders.append([rr, Edge.MinY, 1])
                        continue
                    elif ra == "_":
                        borders.append([rr, Edge.MinY, 0])
                        continue

                    _ra = ra.replace("_", "")
                    if ra.startswith("__"):
                        borders.append([_ra, Edge.MaxY, 1])
                    elif ra.startswith("_"):
                        borders.append([_ra, Edge.MaxY, 0])
                    elif ra.endswith("__"):
                        borders.append([_ra, Edge.MinY, 1])
                    elif ra.endswith("_"):
                        borders.append([_ra, Edge.MinY, 0])
                    
                    ra = _ra
                    if last_area and last_area == ra:
                        _keyed[ra] = _keyed[ra] = union_rect(_keyed[ra], cells[jdx])
                    else:
                        try:
                            _keyed[ra] = cells[jdx]
                        except IndexError:
                            print("-------------------")
                            print(self.columns)
                            print(self.rows)
                            print(self.areas)
                            print(ra)
                            raise Exception("Invalid")
                    last_area = ra
                    jdx += 1
                for k, v in _keyed.items():
                    if k not in keyed:
                        keyed[k] = v
                    else:
                        keyed[k] = union_rect(keyed[k], v)
        if areas:
            if borders:
                b = []
                for a, edge, weight in borders:
                    if isinstance(a, Rect):
                        b.append([a, edge, weight])
                    else: 
                        b.append([keyed[a], edge, weight])
            else:
                b = []

            if True:
                for k, v in keyed.items():
                    x, y, w, h = v
                    if (not float(x).is_integer()
                        or not float(y).is_integer()
                        or not float(w).is_integer()
                        or not float(h).is_integer()):
                        print(">>> FLOAT RECT:::", k, v, "///", r)
            return keyed, b
        else:
            return _grid, []


import skia, tempfile
from subprocess import run
from functools import reduce
from random import Random, randint
from pathlib import Path

from fontTools.svgLib import SVGPath
from fontTools.misc.transform import Transform

from coldtype.color import normalize_color, bw
from coldtype.pens.skiapen import SkiaPen
from coldtype.pens.datpen import DATPen

SKIA_CONTEXT = None

class Skfi():
    @staticmethod
    def contrast_cut(mp=127, w=5):
        ct = bytearray(256)
        for i in range(256):
            if i < mp - w:
                ct[i] = 0
            elif i < mp:
                ct[i] = int((255.0/2)*(1-(mp-i)/w))
            elif i == mp:
                ct[i] = 127
            elif i < mp + w:
                ct[i] = int(127+(255.0/2)*((i-mp)/w))
            else:
                ct[i] = 255
        return ct

    @staticmethod
    def as_filter(lut, a=1, r=0, g=0, b=0):
        args = [lut if x else None for x in [a, r, g, b]]
        return skia.TableColorFilter.MakeARGB(*args)

    @staticmethod
    def compose(*filters):
        # TODO check if ColorFilter or something else?
        return reduce(lambda acc, el: skia.ColorFilters.Compose(el, acc) if acc else el, reversed(filters), None)

    @staticmethod
    def fill(color):
        r, g, b, a = color
        return skia.ColorFilters.Matrix([
            0, 0, 0, 0, r,
            0, 0, 0, 0, g,
            0, 0, 0, 0, b,
            0, 0, 0, a, 0,
        ])

    @staticmethod
    def blur(blur):
        try:
            xblur, yblur = blur
        except:
            xblur, yblur = blur, blur

        return skia.BlurImageFilter.Make(xblur, yblur)

    @staticmethod
    def improved_noise(e, xo=0, yo=0, xs=1, ys=1, base=1):
        noise = skia.PerlinNoiseShader.MakeImprovedNoise(0.015, 0.015, 3, base)
        matrix = skia.Matrix()
        matrix.setTranslate(e*xo, e*yo)
        #matrix.setRotate(45, 0, 0)
        matrix.setScaleX(xs)
        matrix.setScaleY(ys)
        return noise.makeWithLocalMatrix(matrix)


# CHAINABLES

def spackle(xo=None, yo=None,
    xs=0.85, ys=0.85, base=None,
    cut=240, cutw=5
    ):
    r1 = Random()
    r1.seed(0)

    if not xo:
        xo = r1.randint(-500, 500)
    if not yo:
        yo = r1.randint(-500, 500)
    if base is None:
        base = randint(0, 5000)
    
    def _spackle(pen):
        return (pen.f(1)
            .attr(skp=dict(
                Shader=(Skfi.improved_noise(1,
                    xo=xo, yo=yo, xs=xs, ys=ys, base=base)
                    .makeWithColorFilter(Skfi.compose(
                        Skfi.fill(bw(1)),
                        Skfi.as_filter(Skfi.contrast_cut(cut, cutw)),
                        skia.LumaColorFilter.Make()))))))
    return _spackle


def fill(c):
    c = normalize_color(c)
    def _fill(pen):
        return pen.attr(skp=dict(
            ColorFilter=Skfi.fill(c)))
    return _fill


def potrace(rect, poargs=[], invert=True):
    from PIL import Image

    pc = SkiaPen
    ctx = SKIA_CONTEXT

    def _potrace(pen):
        img = pc.Precompose(pen, rect, context=ctx)
        pilimg = Image.fromarray(img.convert(alphaType=skia.kUnpremul_AlphaType))
        binpo = Path("bin/potrace")
        if not binpo.exists():
            binpo = Path(__file__).parent.parent.parent / "bin/potrace"

        with tempfile.NamedTemporaryFile(prefix="coldtype_tmp", suffix=".bmp") as tmp_bmp:
            pilimg.save(tmp_bmp.name)
            rargs = [str(binpo), "-s"]
            if invert:
                rargs.append("--invert")
            rargs.extend([str(x) for x in poargs])
            rargs.extend(["-o", "-", "--", tmp_bmp.name])
            if False:
                print(">>>", " ".join(rargs))
            result = run(rargs, capture_output=True)
            if False:
                print(result)
            t = Transform()
            t = t.scale(0.1, 0.1)
            svgp = SVGPath.fromstring(result.stdout, transform=t)
            if False:
                print(svgp)
            dp = DATPen()
            svgp.draw(dp)
            return dp.f(0)
    return _potrace


def precompose(rect,
    placement=None,
    opacity=1,
    scale=1
    ):
    def _precompose(pen):
        img = SkiaPen.Precompose(pen, rect,             
            context=SKIA_CONTEXT,
            scale=scale,
            disk=False)    
        return (DATPen()
            .rect(placement or rect)
            .img(img, (placement or rect), False, opacity)
            .f(None))
    return _precompose


def rasterized(rect, scale=1):
    """
    Same as precompose but returns the Image created rather
    than setting that image as the attr-image of this pen
    """
    def _rasterized(pen):
        return SkiaPen.Precompose(pen, rect, scale=scale, context=SKIA_CONTEXT, disk=False)
    return _rasterized, dict(returns=skia.Image)


def mod_pixels(rect, scale=0.1, mod=lambda rgba: None):
    import PIL.Image
    
    def _mod_pixels(pen):
        raster = pen.ch(rasterized(rect, scale=scale))
        pi = PIL.Image.fromarray(raster, "RGBa")
        for x in range(pi.width):
            for y in range(pi.height):
                r, g, b, a = pi.getpixel((x, y))
                res = mod((r, g, b, a))
                if res:
                    pi.putpixel((x, y), tuple(res))
        out = skia.Image.frombytes(pi.convert('RGBA').tobytes(), pi.size, skia.kRGBA_8888_ColorType)
        return (DATPen()
            .rect(rect)
            .img(out, rect, False)
            .f(None))
    
    return _mod_pixels


def phototype(rect,
    blur=5,
    cut=127,
    cutw=3,
    fill=1,
    rgba=[0, 0, 0, 1],
    luma=True
    ):
    def _phototype(pen):
        r, g, b, a = rgba

        first_pass = dict(ImageFilter=Skfi.blur(blur))
        
        if luma:
            first_pass["ColorFilter"] = skia.LumaColorFilter.Make()

        cut_filters = [
            Skfi.as_filter(
                Skfi.contrast_cut(cut, cutw),
                a=a, r=r, g=g, b=b)]
            
        if fill is not None:
            cut_filters.append(Skfi.fill(normalize_color(fill)))

        return (pen
            .ch(precompose(rect))
            .attr(skp=first_pass)
            .ch(precompose(rect))
            .attr(skp=dict(
                ColorFilter=Skfi.compose(*cut_filters))))
    return _phototype


def color_phototype(rect,
    blur=5,
    cut=127,
    cutw=15,
    rgba=[1, 1, 1, 1]
    ):
    return phototype(rect, blur, 255-cut, cutw, fill=None, rgba=rgba, luma=False)


import math
from coldtype.geometry.edge import Edge
from fontTools.misc.arrayTools import sectRect, unionRect


MINYISMAXY = False

#https://stackoverflow.com/questions/20677795/how-do-i-compute-the-intersection-point-of-two-lines

def line_intersection(line1, line2):
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return x, y

def calc_vector(point1, point2):
    x1, y1 = point1
    x2, y2 = point2
    dx = x2 - x1
    dy = y2 - y1
    return dx, dy


def calc_angle(point1, point2):
    dx, dy = calc_vector(point1, point2)
    return math.atan2(dy, dx)


def polar_coord(xy, angle, distance):
    x, y = xy
    nx = x + (distance * math.cos(angle))
    ny = y + (distance * math.sin(angle))
    return nx, ny


def centered_square(rect):
    x, y, w, h = rect
    if w > h:
        return [x + (w - h) / 2, y, h, h]
    else:
        return [x, y + (h - w) / 2, w, w]


def perc_to_pix(rect, amount, edge):
    """
    perc(entage) to pix(els) — where the percentage is a decimal between 0 and 1 — cannot be 0 or 1
    """
    x, y, w, h = rect
    if amount <= 1.0:
        d = h if edge == Edge.MinY or edge == Edge.MaxY or edge == Edge.CenterY else w
        if amount < 0:
            return d + amount
        else:
            return d * amount  # math.floor
    else:
        return amount


def divide(rect, amount, edge, forcePixel=False):
    x, y, w, h = rect
    if not forcePixel:
        amount = perc_to_pix(rect, amount, edge)

    if edge == Edge.MaxY:
        if MINYISMAXY:
            return [x, y, w, amount], [x, y + amount, w, h - amount]
        else:
            return [x, y + h - amount, w, amount], [x, y, w, h - amount]
    elif edge == Edge.MinY:
        if MINYISMAXY:
            return [x, y + h - amount, w, amount], [x, y, w, h - amount]
        else:
            return [x, y, w, amount], [x, y + amount, w, h - amount]
    elif edge == Edge.MinX:
        return [x, y, amount, h], [x + amount, y, w - amount, h]
    elif edge == Edge.MaxX:
        return [x + w - amount, y, amount, h], [x, y, w - amount, h]
    elif edge == Edge.CenterX:
        lw = (w - amount) / 2
        return [x, y, lw, h], [x + lw, y, amount, h], [x + lw + amount, y, lw, h]
    elif edge == Edge.CenterY:
        lh = (h - amount) / 2
        return [x, y, w, lh], [x, y + lh, w, amount], [x, y + lh + amount, w, lh]


def subdivide(rect, count, edge, forcePixel=False):
    r = rect
    subs = []
    if hasattr(count, "__iter__"):
        amounts = count
        i = len(amounts) + 1
        a = 0
        while i > 1:
            s, r = divide(r, amounts[a], edge, forcePixel=forcePixel)
            subs.append(s)
            i -= 1
            a += 1
        subs.append(r)
        return subs
    else:
        i = count
        while i > 1:
            s, r = divide(r, 1/i, edge, forcePixel=forcePixel)
            subs.append(s)
            i -= 1
        subs.append(r)
        return subs


def pieces(rect, amount, edge):
    x, y, w, h = rect
    d = w
    if edge == Edge.MaxX or edge == Edge.MaxY:
        d = h
    fit = math.floor(d / amount)
    return subdivide(rect, fit, edge)


def take(rect, amount, edge, forcePixel=False):
    if edge == Edge.CenterX or edge == Edge.CenterY:
        _, r, _ = divide(rect, amount, edge, forcePixel=forcePixel)
        return r
    else:
        r, _ = divide(rect, amount, edge, forcePixel=forcePixel)
        return r


def subtract(rect, amount, edge):
    _, r = divide(rect, amount, edge)
    return r


def drop(rect, amount, edge):
    return subtract(rect, amount, edge)


def inset(rect, dx, dy):
    x, y, w, h = rect
    return [x + dx, y + dy, w - (dx * 2), h - (dy * 2)]


def offset(rect, dx, dy):
    x, y, w, h = rect
    if MINYISMAXY:
        return [x + dx, y - dy, w, h]
    else:
        return [x + dx, y + dy, w, h]


def expand(rect, amount, edge):
    x, y, w, h = rect
    if edge == Edge.MinX:
        w += amount
        x -= amount
    elif edge == Edge.MaxX:
        w += amount
    elif edge == Edge.MinY:
        y -= amount
        h += amount
    elif edge == Edge.MaxY:
        h += amount
    return [x, y, w, h]


def centerpoint(rect):
    x, y, w, h = rect
    return [x + w/2, y + h/2]


def add(rect_a, rect_b):
    # TODO better/correct implementation!
    ax, ay, aw, ah = rect_a
    bx, by, bw, bh = rect_b
    return [
        min(ax, bx),
        min(ay, by),
        aw + bw,
        ah
    ]


def scale(rect, s, x_edge=Edge.CenterX, y_edge=Edge.CenterY):
    """
    Only a partial implementation atm
    """
    x, y, w, h = rect
    return [x * s, y * s, w * s, h * s]


def edgepoints(rect, edge):
    x, y, w, h = rect
    if edge == Edge.MaxY:
        if MINYISMAXY:
            return (x, y), (x + w, y)
        else:
            return (x, y + h), (x + w, y + h)
    elif edge == Edge.MinY:
        if MINYISMAXY:
            return (x, y + h), (x + w, y + h)
        else:
            return (x, y), (x + w, y)
    elif edge == Edge.MinX:
        return (x, y), (x, y + h)
    elif edge == Edge.MaxX:
        return (x + w, y), (x + w, y + h)
    elif edge == Edge.CenterX:
        return (x + w/2, y), (x + w/2, y + h)
    elif edge == Edge.CenterY:
        return (x, y + h/2), (x + w, y + h/2)


import math

def sine(r, periods):
    """Sine-wave primitive"""
    def _record(pen):
        dp = type(pen)()
        pw = r.w / periods
        p1 = r.point("SW")
        end = r.point("SE")
        dp.moveTo(p1)
        done = False
        up = True
        while not done:
            h = r.h if up else -r.h
            c1 = p1.offset(pw/2, 0)
            c2 = p1.offset(pw/2, h)
            p2 = p1.offset(pw, h)
            dp.curveTo(c1, c2, p2)
            p1 = p2
            if p1.x >= end.x:
                done = True
            else:
                done = False
            up = not up
        pen.record(dp)
    return _record
    

def standingwave(r, periods, direction=1):
    """Standing-wave primitive"""
    def _record(pen):
        dp = type(pen)()

        blocks = r.subdivide(periods, "minx")
        for idx, block in enumerate(blocks):
            n, e, s, w = block.take(1, "centery").cardinals()
            if idx == 0:
                dp.moveTo(w)
            if direction == 1:
                if idx%2 == 0:
                    dp.lineTo(n)
                else:
                    dp.lineTo(s)
            else:
                if idx%2 == 0:
                    dp.lineTo(s)
                else:
                    dp.lineTo(n)
            if idx == len(blocks) - 1:
                dp.lineTo(e)
        dp.endPath().smooth()
        dp.value = dp.value[:-1]
        dp.endPath()
        pen.record(dp)
    return _record


def polygon(sides, rect):
    """Polygon primitive; WIP"""
    def _record(pen):
        radius = rect.square().w / 2
        c = rect.center()
        one_segment = math.pi * 2 / sides
        points = [(math.sin(one_segment * i) * radius, math.cos(one_segment * i) * radius) for i in range(sides)]
        dp = type(pen)()
        points.reverse()
        dp.moveTo(points[0])
        for p in points[1:]:
            dp.lineTo(p)
        dp.closePath()
        dp.align(rect)
        pen.record(dp)
    return _record


import json
from pathlib import Path

# import extra stuff for convenience in user code
from coldtype.time.sequence import Marker, ClipTrack, Clip, Sequence, ClipGroup, ClipGroupPens


VIDEO_OFFSET = 86313 # why is this?

group_pens_cache = {}

def to_frames(seconds, fps):
    frames = int(round(float(seconds)*fps))
    if frames >= VIDEO_OFFSET:
        frames -= VIDEO_OFFSET
    return frames


class PremiereTimeline(Sequence):
    __name__ = "Premiere"

    def __init__(self, path, storyboard=None, duration_override=None, workarea_track=0):
        self.path = path

        json_path = path if isinstance(path, Path) else Path(path).expanduser()

        try:
            jsondata = json.loads(json_path.read_text())
        except:
            print(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
            print(json_path.read_text())
            raise Exception("NOPE")
            jsondata = {}

        meta = jsondata.get("metadata")        
        fps = 1 / meta.get("frameRate")
        duration = int(round(int(meta.get("duration"))/int(meta.get("timebase"))))

        if duration_override:
            duration = duration_override

        self.start = 0
        self.end = duration

        tof = lambda s: int(round(float(s)*fps))
        
        cti = tof(meta.get("cti"))
        self.cti = cti

        _storyboard = []
        _storyboard.append(self.cti)
        for m in jsondata.get("storyboard"):
            f = tof(m.get("start"))
            if f not in _storyboard:
                _storyboard.append(f)
        
        if storyboard:
            _storyboard = storyboard
        
        workareas = []
        workareas.append(range(max(0, tof(meta.get("inPoint"))), tof(meta.get("outPoint"))+1))
        self.workareas = workareas

        tracks = []
        for tidx, track in enumerate(jsondata.get("tracks")):
            markers = []
            for marker in track.get("markers"):
                markers.append(Marker(
                    to_frames(marker.get("start"), fps),
                    to_frames(marker.get("end"), fps),
                    marker))
            
            clips = []
            for cidx, clip in enumerate(track.get("clips")):
                clips.append(Clip(
                    clip.get("name"),
                    to_frames(clip.get("start"), fps),
                    to_frames(clip.get("end"), fps),
                    idx=cidx,
                    track=tidx))
            
            tracks.append(ClipTrack(self, clips, markers))
        
        super().__init__(duration, fps, _storyboard, tracks, workarea_track=workarea_track)


import contextlib
import drawBot as db
from coldtype.geometry import Point, Line, Rect
from coldtype.pens.drawbotpen import DrawBotPen
from coldtype.pens.draftingpen import DraftingPen
from coldtype.pens.draftingpens import DraftingPens
from coldtype.text.reader import StyledString, Style, Font
from coldtype.text.composer import StSt
from coldtype.color import hsl, bw
from coldtype.time import Frame
from pathlib import Path

def dbdraw(p:DraftingPen):
    p.cast(DrawBotPen).draw()
    return p

def tobp(p:DraftingPen):
    bp = db.BezierPath()
    p.replay(bp)
    return bp

def dbdraw_with_filters(rect:Rect, filters):
    def _draw_call(p:DraftingPen):
        p.cast(DrawBotPen).draw_with_filters(rect, filters)
        return p
    return _draw_call

def page_rect() -> Rect:
    return Rect(db.width(), db.height())

@contextlib.contextmanager
def new_page(r:Rect=Rect(1000, 1000)):
    _r = Rect(r)
    db.newPage(*_r.wh())
    yield _r

@contextlib.contextmanager
def new_drawing(rect:Rect=Rect(1000, 1000), count=1, save_to=None):
    db.newDrawing()
    for idx in range(0, count):
        with new_page(rect) as r:
            yield idx, r
    if save_to:
        db.saveImage(str(save_to))
    db.endDrawing()

def pdfdoc(fn, path, frame_class=Frame):
    db.newDrawing()
    r = fn.rect
    w, h = r.wh()
    for idx in range(0, fn.duration):
        print(f"Saving page {idx}...")
        db.newPage(w, h)
        if frame_class:
            fn.func(frame_class(idx, fn))
        else:
            fn.func(r)
    pdf_path = Path(path)
    pdf_path.parent.mkdir(exist_ok=True)
    db.saveImage(str(pdf_path))
    print("Saved pdf", str(pdf_path))
    db.endDrawing()


import re
from enum import Enum
from coldtype.time.easing import ease
from coldtype.time.timeable import Timeable


class ClipType(Enum):
    ClearScreen = "ClearScreen"
    NewLine = "NewLine"
    GrafBreak = "GrafBreak"
    Blank = "Blank"
    Isolated = "Isolated"
    JoinPrev = "JoinPrev"
    Meta = "Meta"


class ClipFlags(Enum):
    FadeIn = "FadeIn"
    FadeOut = "FadeOut"


class Clip(Timeable):
    def __init__(self, text, start, end, idx=None, track=0):
        self.idx = idx
        self.input_text = str(text)
        self.text = text
        self.start = start
        self.end = end

        self.styles = []
        self.style_clips = []
        self.position = 1
        self.joined = False
        self.joinPrev = None
        self.joinNext = None
        self.track = track
        self.group = None
        self.blank = False
        self.blank_height = 20
        self.inline_styles = []
        self.inline_data = {}
        self.flags = {}
        self.type = ClipType.Isolated
        self.symbol = None
        self.symbol_position = 0

        if self.text.startswith("^$"):
            symbol, rest = self.text.split("|")
            self.symbol = symbol[2:]
            self.symbol_position = -2
            self.text = rest
        elif self.text.startswith("^"):
            symbol, rest = self.text.split("|")
            self.symbol = symbol[1:]
            self.symbol_position = -1
            self.text = rest
        elif self.text.startswith("$"):
            symbol, rest = self.text.split("|")
            self.symbol = symbol[1:]
            self.symbol_position = +1
            self.text = rest

        if self.text.startswith("*"):
            self.text = self.text[1:]
            self.type = ClipType.ClearScreen
        elif self.text.startswith("≈"):
            self.text = self.text[1:]
            self.type = ClipType.NewLine
        elif self.text.startswith("¶"):
            self.text = self.text[1:]
            self.type = ClipType.GrafBreak
        elif self.text.startswith("+"):
            self.text = self.text[1:]
            self.type = ClipType.JoinPrev
        elif self.text.startswith("µ:"):
            self.type = ClipType.Meta
            self.text = self.text[2:]
        
        parts = self.text.split(":")
        inline_style_marker = parts.index("ß") if "ß" in parts else -1
        inline_data_marker = parts.index("∂") if "∂" in parts else -1

        if inline_style_marker > -1:
            self.inline_styles = parts[inline_style_marker+1].split(",")
        if inline_data_marker > -1:
            qs = [q.split("=") for q in parts[inline_data_marker+1].split("&")]
            for k, v in qs:
                self.inline_data[k] = eval(v)
        
        self.text = parts[0]
        self.text = self.text.replace("{colon}", ":")

        #if "ß" in self.text:
        #    parts = self.text.split(":")
        #    self.text = parts[0]
        #    self.inline_styles = parts[-1].split(",")
        
        #self.text = self.text.replace("§", "\u200b")
        if self.text.startswith("§"):
            #self.text = "\u200b"
            self.text = self.text[1:]
            #self.type = ClipType.JoinPrev
            self.type = ClipType.NewLine
            self.blank = True
            try:
                self.blank_height = float(self.text.strip())
            except:
                pass
    
        if self.text.startswith("∫"):
            self.text = ""
            self.blank = True
        
        if not self.text:
            self.blank = True
            self.text = ""
        
        if self.text.startswith("ƒ"):
            r = r"^([0-9]+)ƒ"
            self.text = self.text[1:]
            value = 3
            match = re.match(r, self.text)
            if match:
                value = int(match[1])
                self.text = re.sub(r, "", self.text)
            self.flags[ClipFlags.FadeIn] = value
        
        elif self.text.endswith("ƒ"):
            self.flags[ClipFlags.FadeOut] = 3
        
        self.original_text = str(self.text)
    
    def addJoin(self, clip, direction):
        if direction == -1:
            self.joinPrev = clip
        elif direction == 1:
            self.joinNext = clip
    
    def joinStart(self):
        if self.joinPrev:
            return self.joinPrev.joinStart()
        else:
            return self.start
    
    def joinEnd(self):
        if self.joinNext:
            return self.joinNext.joinEnd()
        else:
            return self.end
    
    def textForIndex(self, index):
        txt = self.text
        try:
            txt = self.text.split("/")[index]
        except:
            txt = self.text.split("/")[-1]
        return txt, self.addSpace
    
    def style_matching(self, txt):
        try:
            return next(c for c in self.style_clips if txt in c.text)
        except StopIteration:
            return None

    def ftext(self):
        txt = self.text
        if self.type == ClipType.Isolated:
            return " " + txt
        else:
            return txt
    
    def fade(self, default_fade=5):
        if ClipFlags.FadeIn in self.flags:
            return self.flags[ClipFlags.FadeIn]
        else:
            return default_fade
    
    def fadeIn(self, fi, easefn="seio", fade_length=None, start=0):
        if ClipFlags.FadeIn in self.flags or fade_length:
            if fade_length:
                fade = fade_length
            else:
                fade = self.flags[ClipFlags.FadeIn]
            
            if start == 0:
                start_frame = self.start
            elif start == -1:
                start_frame = self.start - fade
            
            if fi < start_frame:
                return -1
            elif fi > start_frame + fade:
                return 1
            else:
                fv = (fi - start_frame) / fade
                a, _ = ease(easefn, fv)
                return a
        return -1
    
    def __repr__(self):
        return "<Clip:({:s}/{:04d}/{:04d}\"{:s}\")>".format([" -1", "NOW", " +1"][self.position+1], self.start, self.end, self.text)